
import re
import json

import pytest

from glom import glom, S, Val, T, A, Fill, Ref, Coalesce, STOP, Switch
from glom.matching import (
    Match, M, MatchError, TypeMatchError, And, Or, Not,
    Optional, Required, Regex)
from glom.core import Auto, SKIP, Ref

try:
    unicode
except NameError:
    unicode = str



def _chk(spec, good_target, bad_target):
    glom(good_target, spec)
    with pytest.raises(MatchError):
        glom(bad_target, spec)

def test_basic():
    _chk(Match(1), 1, 2)
    _chk(Match(int), 1, 1.0)
    # test unordered sequence comparisons
    _chk(Match([int]), [1], ["1"])
    _chk(Match({int}), {1}, [1])
    _chk(Match(frozenset({float})), frozenset({}), frozenset({"1"}))
    _chk(Match(len), [1], [])
    with pytest.raises(MatchError):
        glom(None, Match(len))
    with pytest.raises(MatchError):
        glom([1], Match([]))  # empty shouldn't match
    glom({"a": 1, "b": 2}, Match({str: int}))
    glom(2, M == 2)
    glom(int, M == int)
    glom(1.0, M > 0)
    glom(1.0, M >= 1)
    glom(1.0, M < 2)
    glom(1.0, M <= 1)
    glom(1.0, M != None)
    glom(1.0, (M > 0) & float)
    glom(1.0, (M > 100) | float)

    assert Match(('a', 'b')).matches(('a', 'b', 'c')) is False

    # test idiom for enum
    with pytest.raises(MatchError):
        glom("c", Match("a"))
    glom("c", Not(Match("a")))
    with pytest.raises(MatchError):
        glom("c", Match(Or("a", "b")))

    with pytest.raises(ValueError):
        And()

    with pytest.raises(TypeError):
        And('a', bad_kwarg=True)

    with pytest.raises(ValueError):
        Or()

    with pytest.raises(TypeError):
        Or('a', bad_kwarg=True)


    _chk(Match(Or("a", "b")), "a", "c")
    glom({None: 1}, Match({object: object}))
    _chk(Match((int, str)), (1, "cat"), (1, 2))
    with pytest.raises(MatchError):
        glom({1: 2}, Match({(): int}))
    with pytest.raises(MatchError):
        glom(1, Match({}))
    Match(M > 0).verify(1.0)

    assert Match(M).matches(False) is False
    assert Match(M).matches(True) is True


def test_match_expressions():
    assert glom(1, M == M) == 1
    assert glom(1, M == 1) == 1
    assert glom(1, M >= 1) == 1
    assert glom(1, M <= 1) == 1
    with pytest.raises(MatchError):
        glom(1, M > 1)
    with pytest.raises(MatchError):
        glom(1, M < 1)
    with pytest.raises(MatchError):
        glom(1, M != 1)


def test_defaults():
    assert glom(1, Match(2, default=3)) == 3
    assert glom(1, Or(M == 2, default=3)) == 3
    assert glom(1, And(M == 2, default=3)) == 3


def test_match_default():
    default = []
    res = glom(None, Match(list, default=default))
    assert res == default
    default = []
    res = glom(None, Match(list, default=Val(default)))
    assert res is default


def test_double_wrapping():
    for outer in (Required, Optional):
        with pytest.raises(TypeError):
            outer(Optional('key'))

    with pytest.raises(ValueError):
        Required('key')

    return


def test_sets():
    with pytest.raises(MatchError):
        glom({1}, Match({}))
    with pytest.raises(MatchError):
        glom(frozenset([1]), Match(frozenset()))


def test_m_call_match():
    """test that M __call__ can be used to wrap a subspec for comparison"""
    target = {}
    target['a'] = target
    assert glom(target, M == M(T['a'])) == target
    assert glom(target, M(T['a']) == M) == target
    assert repr(M(T['a'])) == "M(T['a'])"

    with pytest.raises(TypeError):
        M('failure')  # TODO: may change in future, see TODO in M.__call__

    with pytest.raises(MatchError):
        glom({'a': False}, M(T['a']))

    # let's take the operators for a spin
    valid_target_spec_pairs = [({'a': 1, 'b': 2}, M(T['a']) < M(T['b'])),
                               ({'a': 2, 'b': 1}, M(T['a']) > M(T['b'])),
                               ({'a': 2, 'b': 1}, M(T['a']) != M(T['b'])),
                               ({'a': 2, 'b': 1}, M(T['a']) >= M(T['b'])),
                               ({'a': 2, 'b': 2}, M(T['a']) >= M(T['b'])),
                               ({'a': 2, 'b': 2}, M(T['a']) <= M(T['b'])),
                               ({'a': 1, 'b': 2}, M(T['a']) <= M(T['b']))]

    for target, spec in valid_target_spec_pairs:
        assert glom(target, spec) is target

    return


def test_and_or_reduction():
    and_spec = And(T['a'], T['b']) & T['c']

    assert repr(and_spec) == "And(T['a'], T['b'], T['c'])"

    or_spec = Or(T['a'], T['b']) | T['c']

    assert repr(or_spec) == "Or(T['a'], T['b'], T['c'])"


def test_precedence():
    """test corner cases of dict key precedence"""
    glom({(0, 1): 3},
        Match({
            (0, 1): Val(1),  # this should match
            (0, int): Val(2),  # optional
            (0, M == 1): Val(3),  # optional
        })
    )
    with pytest.raises(ValueError):
        Optional(int)  # int is already optional so not valid to wrap


def test_cruddy_json():
    _chk(
        Match({'int_id?': Auto((int, (M > 0)))}),
        {'int_id?': '1'},
        {'int_id?': '-1'})
    # embed a build
    squished_json = Match({
        'smooshed_json': Auto(
            (json.loads, Match({
                'sub': Auto((json.loads, M == 1))})))
        })
    glom({'smooshed_json': json.dumps({'sub': json.dumps(1)})}, squished_json)


def test_pattern_matching():
    pattern_matcher = Or(
        And(Match(1), Val('one')),
        And(Match(2), Val('two')),
        And(Match(float), Val('float'))
        )
    assert glom(1, pattern_matcher) == 'one'
    assert glom(1.1, pattern_matcher) == 'float'

    # obligatory fibonacci

    fib = (M > 2) & (lambda n: glom(n - 1, fib) + glom(n - 2, fib)) | T

    assert glom(5, fib) == 8

    factorial = (
        lambda t: t + 1, Ref('fact', (
            lambda t: t - 1,
            (M == 0) & Fill(1) |
            (S(r=Ref('fact')),
             S, lambda s: s['r'] * s[T]))))

    assert glom(4, factorial) == 4 * 3 * 2


def test_examples():
    assert glom(8, (M > 7) & Val(7)) == 7
    assert glom(range(10), [(M > 7) & Val(7) | T]) == [0, 1, 2, 3, 4, 5, 6, 7, 7, 7]
    assert glom(range(10), [(M > 7) & Val(SKIP) | T]) == [0, 1, 2, 3, 4, 5, 6, 7]


def test_reprs():
    repr(M)
    repr(M == 1)
    repr(M | M == 1)
    repr(M & M == 1)
    repr(~M)
    repr(And(1, 2))
    repr(Or(1, 2))
    repr(Not(1))
    repr(MatchError("uh oh"))
    repr(TypeMatchError("uh oh {0}", dict))
    assert repr(And(M == 1, float)) == "(M == 1) & float"
    assert repr(eval(repr(And(M == 1, float)))) == "(M == 1) & float"

    assert repr(Regex('[ab]')) == "Regex('[ab]')"
    assert repr(Regex('[ab]', flags=1)) == "Regex('[ab]', flags=1)"
    assert 'search' in repr(Regex('[ab]', func=re.search))
    assert repr(And(1)) == 'And(1)'
    assert repr(~And(1)) == 'Not(And(1))'
    assert repr(~Or(M) & Or(M)) == '~(M) & M'
    assert repr(Not(M < 3)) == '~(M < 3)'
    assert repr(~(M < 4)) == '~(M < 4)'
    assert repr(~M | "default") == "~M | 'default'"
    assert repr(Or(M, default=1)) == "Or(M, default=1)"


def test_shortcircuit():
    assert glom(False, Fill(M | "default")) == "default"
    assert glom(True, Fill(M | "default")) == True
    assert glom(True, Fill(M & "default")) == "default"
    with pytest.raises(MatchError):
        glom(False, Fill(M & "default"))
    assert glom(False, ~M) == False
    assert glom(True, Fill(~M | "default")) == "default"


def test_sample():
    """
    test meant to cover a more realistic use
    """
    import datetime

    data = {
        'name': 'item',
        'date_added': datetime.datetime.now(),
        'desc': 'a data item',
        'tags': ['data', 'new'],
    }

    spec = Match({
        'name': str,
        Optional('date_added'): datetime.datetime,
        'desc': str,
        'tags': [str,]})

    def good():
        glom(data, spec)
    def bad():
        with pytest.raises(MatchError):
            glom(data, spec)

    good()  # should match
    del data['date_added']
    good()  # should still match w/out optional
    del data['desc']
    bad()
    data['desc'] = 'a data item'
    data['extra'] = 'will fail on extra'
    bad()
    spec.spec[str] = str  # now extra str-key/str-val are okay
    good()
    data['extra2'] = 2  # but extra str-key/non-str-val are bad
    bad()
    # reset data
    data = {
        'name': 'item',
        'date_added': datetime.datetime.now(),
        'desc': 'a data item',
        'tags': ['data', 'new'],
    }
    del spec.spec[str]
    spec.spec[Required(str)] = str  # now there MUST be at least one str
    bad()
    data['extra'] = 'extra'
    good()


def test_regex():
    assert glom('abc', (Regex('(?P<test>.*)'), S['test'])) == 'abc'
    # test wrong target type failure path
    with pytest.raises(MatchError):
        glom(1, Regex('1'))
    # test invalid arg path
    with pytest.raises(ValueError):
        Regex(1, func="invalid")
    # test explicit re match func and target value failure path
    with pytest.raises(MatchError):
        glom('aabcc', Regex('abc', func=re.match))


def test_ternary():
    assert glom('abc', Match(Or(None, 'abc'))) == 'abc'


def test_sky():
    """test adapted from github.com/shopkick/opensky"""

    def as_type(sub_schema, typ):
        'after checking sub_schema, pass the result to typ()'
        return And(sub_schema, Auto(typ))

    assert glom('abc', as_type(M == 'abc', list)) == list('abc')

    def none_or(sub_schema):
        'allow None or sub_schema'
        return Match(Or(None, sub_schema))

    assert glom(None, none_or('abc')) == None
    assert glom('abc', none_or('abc')) == 'abc'
    with pytest.raises(MatchError):
        glom(123, none_or('abc'))

    def in_range(sub_schema, _min, _max):
        'check that sub_schema is between _min and _max'
        return Match(And(sub_schema, _min < M, M < _max))
        # TODO: make _min < M < _max work

    assert glom(1, in_range(int, 0, 2))
    with pytest.raises(MatchError):
        glom(-1, in_range(int, 0, 2))

    def default_if_none(sub_schema, default_factory):
        return Or(
            And(M == None, Auto(lambda t: default_factory())), sub_schema)

    assert glom(1, default_if_none(T, list)) == 1
    assert glom(None, default_if_none(T, list)) == []

    def nullable_list_of(*items):
        return default_if_none(Match(list(items)), list)

    assert glom(None, nullable_list_of(str)) == []
    assert glom(['a'], nullable_list_of(str)) == ['a']
    with pytest.raises(MatchError):
        glom([1], nullable_list_of(str))


def test_clamp():
    assert glom(range(10), [(M < 7) | Val(7)]) == [0, 1, 2, 3, 4, 5, 6, 7, 7, 7]
    assert glom(range(10), [(M < 7) | Val(SKIP)]) == [0, 1, 2, 3, 4, 5, 6]


def test_json_ref():
    assert glom(
        {'a': {'b': [0, 1]}},
        Ref('json',
            Match(Or(
                And(dict, {Ref('json'): Ref('json')}),
                And(list, [Ref('json')]),
                And(0, Val(None)),
                object)))) == {'a': {'b': [None, 1]}}


def test_nested_struct():
    """adapted from use case"""
    import json

    _json = lambda spec: Auto((json.loads, _str_json, Match(spec)))

    _str_json = Ref('json',
        Match(Or(
            And(dict, {Ref('json'): Ref('json')}),
            And(list, [Ref('json')]),
            And(type(u''), Auto(str)),
            object)))

    rule_spec = Match({
        'rule_id': Or('', Regex(r'\d+')),
        'rule_name': str,
        'effect': Or('approve', 'custom_approvers'),
        'rule_approvers': _json([{'pk': int, 'level': int}]),
        'rule_data': _json([  # list of condition-objects
            {
                Optional('value', 'null'): _json(
                    Or(None, int, float, str, [int, float, str])),
                'field': Auto(int),  # id of row from FilterField
                'operator': str,  # corresponds to FilterOperator.display_name
            }]),
        Optional('save_as_new', False): Or(str, bool),
    })

    rule = dict(
        rule_id='1',
        rule_name='test rule',
        effect='approve',
        rule_approvers=json.dumps([{'pk': 2, 'level': 1}]),
        rule_data=json.dumps([
            {'value': json.dumps([1, 2]), 'field': 2, 'operator': '>'},
            {'field': 2, 'operator': '=='}])
    )

    glom(rule, rule_spec)
    rule['save_as_new'] = 'true'
    glom(rule, rule_spec)


def test_check_ported_tests():
    """
    Tests ported from Check() to make sure all the functionality has an analogue.
    """
    target = [{'id': 0}, {'id': 1}, {'id': 2}]

    # check that skipping non-passing values works
    assert glom(target, [Coalesce(M(T['id']) == 0, default=SKIP)]) == [{'id': 0}]

    # TODO: should M(subspec, default='') work? I lean no.
    # NB: this is not a very idiomatic use of Match, just brought over for Check reasons
    assert glom(target, [Match({'id': And(int, M == 1)}, default=SKIP)]) == [{'id': 1}]
    assert glom(target, [Match({'id': And(int, M <= 1)}, default=STOP)]) == [{'id': 0}, {'id': 1}]

    # check that stopping chain execution on non-passing values works
    spec = (Or(Match(len), Val(STOP)), T[0])
    assert glom('hello', spec, glom_debug=True) == 'h'
    assert glom('', spec) == ''  # would fail with IndexError if STOP didn't work

    target = [1, u'a']
    assert glom(target, [Match(unicode, default=SKIP)]) == ['a']
    assert glom(target, Match([Or(unicode, int)])) == [1, 'a']

    target = ['1']
    assert glom(target, [(M(T), int)]) == [1]
    assert glom(target, M(T)) == ['1']

    failing_checks = [({'a': {'b': 1}}, {'a': ('a', 'b', Match(str))},
                       '''expected type str, not int'''),  # TODO: bbrepr at least, maybe include path like Check did
                      ({'a': {'b': 1}}, {'a': ('a', Match({'b': str}))},
                       '''expected type str, not int'''),  # TODO: include subspec path ('b')
                      (1, Match(Or(unicode, bool))),
                      (1, Match(unicode)),
                      (1, Match(0)),
                      (1, Match(Or(0, 2))),
                      ('-3.14', Match(lambda x: int(x) > 0)),
                      # ('-3.14', M(lambda x: int(x) > 0)),
                      # TODO: M doesn't behave quite like Match because it's mode-free
    ]

    for fc in failing_checks:
        if len(fc) == 2:
            target, check = fc
            msg = None
        else:
            target, check, msg = fc

        with pytest.raises(MatchError) as exc_info:
            glom(target, check)

        if msg is not None:
            actual_msg = str(exc_info.value)
            assert actual_msg.find(msg) != -1
        assert repr(exc_info.value)

    return


def test_switch():
    data = {'a': 1, 'b': 2}
    cases = [('c', lambda t: 3), ('a', 'a')]
    cases2 = dict(cases)
    assert glom(data, Switch(cases)) == 1
    assert glom(data, Switch(cases2)) == 1
    assert glom({'c': None}, Switch(cases)) == 3
    assert glom({'c': None}, Switch(cases2)) == 3
    assert glom(None, Switch(cases, default=4)) == 4
    assert glom(None, Switch({'z': 26}, default=4)) == 4
    with pytest.raises(MatchError):
    	glom(None, Switch(cases))
    with pytest.raises(ValueError):
    	Switch({})
    with pytest.raises(TypeError):
    	Switch("wrong type")
    assert glom(None, Switch({S(a=1): S['a']})) == 1
    repr(Switch(cases))


def test_nested_dict():
    assert glom({1: 2}, Match({A.t: S.t})) == {1: 1}



"""
Config subsystem
We use a module-level singleton pattern here, where top-level functions and globals
act like the state for a "Manager" class around the internal Config object
"""
import configparser
import dataclasses as dc
import typing as t
import uuid
from contextlib import suppress
from os import getenv
from pathlib import Path
from typing import Optional

import click
from typing_extensions import Self

from .exceptions import InvalidTokenError
from .utils import IN_PYTEST, log

APP_NAME = "datapane"
APP_DIR = Path(getenv("DATAPANE_APP_DIR", click.get_app_dir(APP_NAME)))
APP_DIR.mkdir(parents=True, exist_ok=True)
CONFIG_FILENAME = "config.ini"
LEGACY_CONFIG_FILENAME = "default.yaml"
CONFIG_PATH = Path(getenv("DATAPANE_CONFIG", APP_DIR / CONFIG_FILENAME))
LEGACY_CONFIG_PATH = APP_DIR / LEGACY_CONFIG_FILENAME
CONFIG_SECTION = "default"
PAST_DEFAULT_SERVER = "https://datapane.com"
DEFAULT_SERVER = "https://cloud.datapane.com"
DEFAULT_TOKEN = "TOKEN_HERE"
LATEST_VERSION = 6


# TODO - wrap into a singleton object that includes callable?
@dc.dataclass
class Config:
    """
    Global config read from config file

    Versioning
    - for b/c from before versioning config file, we set version to 1 if doesn't exist
    - else we set it to latest version for a new config via post_init hook
    - only on loading a config file do we run the upgrade function
    """

    server: str = DEFAULT_SERVER
    token: str = DEFAULT_TOKEN
    email: str = ""
    session_id: str = dc.field(default_factory=lambda: uuid.uuid4().hex)
    version: int = 1  # for b/c if version doesn't exist in file, set to latest in post_init hook
    completed_action: bool = False  # only active on first action

    from_file: dc.InitVar[bool] = False

    def __post_init__(self, from_file: bool):
        self.server = self.server.rstrip("/")  # server should be a valid origin
        if not from_file:
            # set to latest if generating config file
            self.version = LATEST_VERSION

    @property
    def is_public(self) -> bool:
        return self.server == DEFAULT_SERVER

    @property
    def is_org(self) -> bool:
        return not self.is_public

    @property
    def is_authenticated(self) -> bool:
        return self.token != DEFAULT_TOKEN  # or bool(self.email)

    @property
    def is_anonymous(self) -> bool:
        return not self.is_authenticated

    # MANAGER functions
    @classmethod
    def load(cls) -> Self:
        """Load config for an environment and set globally"""
        # Read config file
        cls.ensure_config_file()
        parser = configparser.ConfigParser()
        parser.read(CONFIG_PATH)
        if CONFIG_SECTION not in parser:
            log.error(f"Config file at {CONFIG_PATH} does not contain a {CONFIG_SECTION} section")
            raise ValueError(f"Config file does not contain a {CONFIG_SECTION} section")

        # Process config into a new dataclass
        raw_conf: configparser.SectionProxy = parser[CONFIG_SECTION]
        kwargs: t.Dict[str, t.Any] = {
            "from_file": True,
        }
        for field in dc.fields(cls):
            if field.name in raw_conf:
                if field.type is bool:
                    value = raw_conf[field.name].lower() == "true"
                else:
                    value = field.type(raw_conf[field.name])
                kwargs[field.name] = value

        config: Config = Config(**kwargs)
        log.debug(f"Loaded client environment from {CONFIG_PATH}")

        # check if stored file is out of date
        config.upgrade_config_format()

        # set to the global state
        set_config(config)
        return config

    @classmethod
    def create_default(cls) -> None:
        """Create an default config file"""
        # create default file
        _config = Config()
        _config.save()
        log.info(f"Created config file at {CONFIG_PATH}")

    def save(self):
        data = dc.asdict(self)
        parser = configparser.ConfigParser()
        parser[CONFIG_SECTION] = data
        with CONFIG_PATH.open("w") as file:
            parser.write(file)

    def remove(self):
        CONFIG_PATH.unlink()

    @classmethod
    def ensure_config_file(cls) -> None:
        """
        Ensure the config file exists at CONFIG_PATH
        """
        if CONFIG_PATH.exists():
            return

        # Look for YAML from Version 5 and earlier
        yaml_path = LEGACY_CONFIG_PATH
        if yaml_path.exists():
            # Migrate
            log.info(f"Found legacy config file at {yaml_path}")
            CONFIG_PATH.write_text(f"[{CONFIG_SECTION}]\n" + yaml_path.read_text())
            yaml_path.unlink()
            log.info(f"Successfully migrated to {CONFIG_PATH}")

        else:
            # Neither exist - create new
            cls.create_default()

    def upgrade_config_format(self):
        """Handles updating the older config format
        - we default to the oldest version with default values, and upgrade here
        """
        # this isn't tied to a particular config version
        if self.server == PAST_DEFAULT_SERVER:
            self.server = DEFAULT_SERVER

        # migrate older config files
        if self.version in (1, 2, 3):
            # If token exists check still valid and can login, use to get server props
            if self.token and self.token != DEFAULT_TOKEN:
                from ..cloud_api.user import ping

                with suppress(Exception):
                    # get the email for v4 of spec
                    self.email = ping(config=self, cli_login=True, verbose=False)

                # if not self.completed_action:
                #     # we could be on older version, but with a valid token
                #     # but haven't completed an action, so force it
                #     from .analytics import capture
                #
                #     capture("CLI Login", config=self, with_token=True)

            self.version = LATEST_VERSION
            self.save()
        elif self.version in (4, 5):
            self.version = LATEST_VERSION
            self.save()  # trigger removal of _env & _pass from file


# TODO - create a ConfigMgr singleton object?
config: Optional[Config] = None


################################################################################
# MODULE LEVEL INTERFACE
def init(config: t.Optional[Config] = None) -> Config:
    """
    Init an API config
     - this MUST handle being called multiple times and only from the main-thread
    """
    if globals().get("config") is not None:
        log.debug("Reinitialising client config")

    if config:
        set_config(config)
    else:
        config = Config.load()

    return config


def check_get_config() -> Config:
    """Attempt to get a config object, reloading if necessary
    - used when we need a valid API token, e.g. when performing a network action"""
    config = get_config()
    if config.token == DEFAULT_TOKEN:
        # try reinit, as may have ran login in another terminal/subprocess
        _config = init()
        if _config.token == DEFAULT_TOKEN:
            # still don't have a token set, open up the browser and wait for login
            if not IN_PYTEST:
                from datapane.cloud_api import signup

                signup()
                _config = init()
                return _config
            raise InvalidTokenError(
                "Please sign-up and login - if you already have then please restart your Jupyter kernel/Python instance to initialize your new token"
            )

        return _config
    return config


def set_config(c: Optional[Config]):
    global config
    config = c


def get_config() -> Config:
    """Get the current config object, doesn't attempt to re-init the API token"""
    global config
    if config is None:
        raise RuntimeError("Config must be initialised before it can be used")

    return config



# flake8: noqa:F811
import dataclasses as dc
import io
import tarfile
import time
import typing as t
import uuid
from contextlib import contextmanager
from distutils.dir_util import copy_tree
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click
import click_spinner
import importlib_resources as ir
from jinja2 import Environment, FileSystemLoader
from requests import HTTPError
from tabulate import tabulate

from datapane import __rev__, __version__
from datapane import cloud_api as api
from datapane import legacy_apps as apps
from datapane.app import hosting_utils
from datapane.client import log
from datapane.client.utils import _setup_dp_logging
from datapane.common import SDict, dict_drop_empty
from datapane.common.dp_types import URL
from datapane.legacy_apps import config as sc

from . import analytics
from . import config as c
from .exceptions import DPClientError, add_help_text
from .utils import process_cmd_param_vals

EXTRA_OUT: bool = False


# TODO
#  - add info subcommand
#  - convert to use typer (https://github.com/tiangolo/typer) or autoclick
def init(verbosity: int):
    """Init the cmd-line env"""
    c.init()
    _setup_dp_logging(verbosity=verbosity)


@dc.dataclass(frozen=True)
class DPContext:
    """
    Any shared context we want to pass across commands,
    easier to just use globals in general tho
    """


@contextmanager
def api_error_handler(err_msg: str):
    try:
        yield
    except HTTPError as e:
        if EXTRA_OUT:
            log.exception(e)
        else:
            log.error(e)
        failure_msg(err_msg, do_exit=True)


def gen_name() -> str:
    return f"new_{uuid.uuid4().hex}"


def print_table(xs: t.Iterable[SDict], obj_name: str, showindex: bool = True) -> None:
    success_msg(f"Available {obj_name}:")
    print(tabulate(xs, headers="keys", showindex=showindex))


class GlobalCommandHandler(click.Group):
    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except Exception as e:
            analytics.capture("CLI Error", msg=str(e), type=str(type(e)))
            if EXTRA_OUT:
                log.exception(e)
            if isinstance(e, DPClientError):
                failure_msg(str(e))
            else:
                failure_msg(add_help_text(str(e)), do_exit=True)


def recursive_help(cmd, parent=None):  # noqa: ANN001
    ctx = click.core.Context(cmd, info_name=cmd.name, parent=parent)
    print(cmd.get_help(ctx))
    print()
    commands = getattr(cmd, "commands", {})
    for sub in commands.values():
        recursive_help(sub, ctx)


###############################################################################
# Main
@click.group(cls=GlobalCommandHandler)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Increase logging verbosity - can add multiple times",
)
@click.version_option(version=f"{__version__} ({__rev__})")
@click.pass_context
def cli(ctx: click.core.Context, verbose: int):
    """Datapane CLI Tool"""
    global EXTRA_OUT
    EXTRA_OUT = verbose > 0
    init(verbosity=verbose)
    ctx.obj = DPContext()


# @cli.command()
# def dumphelp():
#     recursive_help(cli)


###############################################################################
# Auth
@cli.command()
@click.option("--token", default=None, help="API Token to the Datapane server.")
@click.option("--server", default=c.DEFAULT_SERVER, help="Datapane API Server URL.")
@click.pass_obj
def login(obj: DPContext, token: str, server: str):
    """Login to a server with the given API token."""
    api.login(token, server)
    # click.launch(f"{server}/settings/")


@cli.command()
@click.option("--server", default=c.DEFAULT_SERVER, help="Datapane API Server URL.")
def signup(server: str):
    """Signup and link your account to the Datapane CLI"""
    api.signup(server)


@cli.command()
@click.pass_obj
def logout(obj: DPContext):
    """Logout from the server and reset the config file"""
    api.logout()


@cli.command()
def ping():
    """Check can connect to the server"""
    try:
        api.ping()
    except HTTPError as e:
        log.error(e)


@cli.command()
def hello_world():
    """Create and run an example report, and open in the browser"""
    api.hello_world()


@cli.command()
@click.argument("url", default=None)
@click.option("--execute/--no-execute", default=True)
def template(url: URL, execute: bool):
    """Retrieve and run a template report, and open in the browser

    URL is the location of the template repository. Relative locations can be used for first-party templates, e.g. `dp-template-classifier-dashboard`.
    """
    api.template(url, execute)


###############################################################################
# Files
@cli.group()
def file():
    """Commands to work with Files"""


@file.command()
@click.argument("name")
@click.argument("file", type=click.Path(exists=True))
@click.option("--project")
@click.option("--overwrite", is_flag=True)
def upload(file: str, name: str, project: str, overwrite: bool):
    """Upload a csv or Excel file as a Datapane File"""
    log.info(f"Uploading {file}")
    r = api.File.upload_file(file, name=name, project=project, overwrite=overwrite)
    success_msg(f"Uploaded {click.format_filename(file)} to {r.url}")


@file.command()
@click.argument("name")
@click.option("--project")
@click.argument("file", type=click.Path())
def download(name: str, project: str, file: str):
    """Download file referenced by NAME to FILE"""
    r = api.File.get(name, project=project)
    r.download_file(file)
    success_msg(f"Downloaded {r.url} to {click.format_filename(file)}")


@file.command()
@click.argument("name")
@click.option("--project")
def delete(name: str, project: str):
    """Delete a file"""
    api.File.get(name, project).delete()
    success_msg(f"Deleted File {name}")


@file.command("list")
def file_list():
    """List files"""
    print_table(api.File.list(), "Files")


###############################################################################
# Apps
@cli.group()
def app():
    """Commands to work with Apps"""


def write_templates(scaffold_name: str, context: SDict):
    """Write templates for the given local scaffold (TODO - support git repos)"""
    # NOTE - only supports single hierarchy project dirs
    env = Environment(loader=FileSystemLoader("."))

    def render_file(fname: Path, context: Dict):
        rendered_script = env.get_template(fname.name).render(context)
        fname.write_text(rendered_script)

    # copy the scaffolds into the service
    def copy_scaffold() -> List[Path]:
        dir_path = ir.files("datapane.resources.app_templates") / scaffold_name
        copy_tree(str(dir_path), ".")
        return [Path(x.name) for x in dir_path.iterdir()]

    # run the scripts
    files = copy_scaffold()
    for f in files:
        if f.exists() and f.is_file():
            render_file(f, context=context)


@app.command(name="init")
@click.argument("name", default=lambda: sc.generate_name("Report"))
def app_init(name: str):
    """Initialise a new app project"""
    if sc.DATAPANE_YAML.exists():
        failure_msg("Found existing project, cancelling", do_exit=True)

    sc.validate_name(name)
    _context = dict(name=name)

    write_templates("app", _context)
    success_msg(f"Created dp_app.py for project '{name}', edit as needed and upload")


@app.command()
@click.option("--config", type=click.Path(exists=True))
@click.option("--script", type=click.Path(exists=True))
@click.option("--name")
@click.option("--environment")
@click.option("--project")
@click.option("--overwrite", is_flag=True)
def deploy(
    name: Optional[str],
    script: Optional[str],
    config: Optional[str],
    environment: Optional[str],
    project: Optional[str],
    overwrite: Optional[bool],
):
    """Package and deploy a Python script or Jupyter notebook as a Datapane App bundle"""
    script_path = script and Path(script)
    config_path = config and Path(config)
    init_kwargs = dict(name=name, script=script_path, config_file=config_path, environment=environment, project=project)
    kwargs = dict_drop_empty(init_kwargs, none_only=True)

    # if not (script or config or sc.DatapaneCfg.exists()):
    #     raise DPClientError(f"Not valid project dir")

    dp_cfg = apps.DatapaneCfg.create_initial(**kwargs)
    log.debug(f"Packaging and uploading Datapane project {dp_cfg.name}")

    # start the build process
    with apps.build_bundle(dp_cfg) as sdist:
        if EXTRA_OUT:
            tf: tarfile.TarFile
            log.debug("Bundle from following files:")
            with tarfile.open(sdist) as tf:
                for n in tf.getnames():
                    log.debug(f"  {n}")
        r: api.LegacyApp = api.LegacyApp.upload_pkg(sdist, dp_cfg, overwrite)
        success_msg(f"Uploaded {click.format_filename(str(dp_cfg.script))} to {r.web_url}")


@app.command()  # type: ignore[no-redef]
@click.argument("name")
@click.option("--project")
def download(name: str, project: str):
    """Download app referenced by NAME to FILE"""
    s = api.LegacyApp.get(name, project=project)
    fn = s.download_pkg()
    success_msg(f"Downloaded {s.url} to {click.format_filename(str(fn))}")


@app.command()  # type: ignore[no-redef]
@click.argument("name")
@click.option("--project")
def delete(name: str, project: str):
    """Delete an app"""
    api.LegacyApp.get(name, project).delete()
    success_msg(f"Deleted App {name}")


@app.command("list")
def app_list():
    """List Apps"""
    print_table(api.LegacyApp.list(), "Apps")


@app.command()
@click.option("--parameter", "-p", multiple=True)
@click.option("--cache/--disable-cache", default=True)
@click.option("--wait/--no-wait", default=True)
@click.option("--project")
@click.option("--show-output", is_flag=True, default=False, help="Display the run output")
@click.argument("name")
def run(
    name: str,
    parameter: Tuple[str],
    cache: bool,
    wait: bool,
    project: str,
    show_output: bool,
):
    """Run a report"""
    params = process_cmd_param_vals(parameter)
    log.info(f"Running app with parameters {params}")
    app = api.LegacyApp.get(name, project=project)
    with api_error_handler("Error running app"):
        r = app.run(parameters=params, cache=cache)
    if wait:
        with click_spinner.spinner():  # type: ignore[call-arg]
            while not r.is_complete():
                time.sleep(5)
                r.refresh()
            log.debug(f"Run completed with status {r.status}")
            if show_output:
                click.echo(r.output)
            if r.status == "SUCCESS":
                if r.result:
                    success_msg(f"App result - '{r.result}'")
                if r.report:
                    report = api.CloudReport.by_id(r.report)
                    success_msg(f"Report generated at {report.web_url}")
            else:
                failure_msg(f"App run failed/cancelled\n{r.error_msg}: {r.error_detail}")

    else:
        success_msg(f"App run started, view at {app.web_url}")


###############################################################################
# Reports
@cli.group()
def report():
    """Commands to work with Reports"""


# NOTE - CLI Report creation disabled for now until we have a replacement for Asset
# @report.command()
# @click.argument("name")
# @click.argument("files", type=click.Path(), nargs=-1, required=True)
# def create(files: Tuple[str], name: str):
#     """Create a Report from the provided FILES"""
#     blocks = [api.Asset.upload_file(file=Path(f)) for f in files]
#     r = api.Report(*blocks)
#     r.upload(name=name)
#     success_msg(f"Created Report {r.web_url}")


@report.command(name="init")
@click.argument("name", default=lambda: sc.generate_name("Script"))
@click.option("--format", type=click.Choice(["notebook", "script"]), default="notebook")
def report_init(name: str, format: str):
    """Initialise a new report"""

    if format == "notebook":
        template = "report_ipynb"
    else:
        template = "report_py"

    if len(list(Path(".").glob("dp_report.*"))):
        failure_msg("Found existing project, cancelling", do_exit=True)

    sc.validate_name(name)
    _context = dict(name=name)

    write_templates(template, _context)
    success_msg("Created sample report `dp_report`, edit as needed and run")


@report.command()  # type: ignore[no-redef]
@click.argument("name")
@click.option("--project")
def delete(name: str, project: str):
    """Delete a report"""
    api.CloudReport.get(name, project).delete()
    success_msg(f"Deleted Report {name}")


@report.command("list")
def report_list():
    """List Reports"""
    print_table(api.CloudReport.list(), "Reports")


# NOTE - NYI - disabled
# @report.command()
# @click.argument("name")
# @click.option("--project")
# @click.option("--filename", default="output.html", type=click.Path())
# def render(name: str, project: str, filename: str):
#     """Render a report to a static file"""
#     api.Report.get(name, project=project).render()


#############################################################################
# Environments
@cli.group()
def environment():
    """Commands to work with Environments"""


@environment.command()
@click.argument("name", required=True)
@click.option("--environment", "-env", multiple=True)
@click.option("--docker-image")
@click.option("--project")
@click.option("--overwrite", is_flag=True)
def create(name: str, environment: Tuple[str], docker_image: str, project: str, overwrite: bool):
    """
    Create an environment

    NAME: name of environment
    ENVIRONMENT: environment variables
    DOCKER IMAGE: docker image to be used inside apps
    PROJECT: Project name (optional)
    """
    proccessed_environment = process_cmd_param_vals(environment)
    api.Environment.create(name, proccessed_environment, docker_image, project, overwrite)
    success_msg(f"Created environment: {name}")


@environment.command("list")
def environment_list():
    """List all environments"""
    print_table(api.Environment.list(), "Environments")


@environment.command()
@click.argument("name", required=True)
@click.option("--project")
def get(name: str, project: str):
    """Get environment value using environment name"""
    res = api.Environment.get(name, project=project)
    environment = "\n".join([f"{k}={v}" for k, v in res.environment.items()])
    success_msg(f"Available {res.name}:")
    print(f"\nProject - {res.project}")
    print(f"\nEnvironment Variables-----------\n{environment}")
    print(f"\nDocker Image - {res.docker_image}")


@environment.command()  # type: ignore[no-redef]
@click.argument("name")
@click.option("--project")
def delete(name: str, project: str):
    """Delete an environment using environment name"""
    api.Environment.get(name, project).delete()
    success_msg(f"Deleted environment {name}")


#############################################################################
# Schedules
@cli.group()
def schedule():
    """Commands to work with Schedules"""


@schedule.command()  # type: ignore[no-redef]
@click.option("--parameter", "-p", multiple=True)
@click.argument("name", required=True)
@click.argument("cron", required=True)
@click.option("--project")
def create(name: str, cron: str, parameter: Tuple[str], project: str):
    """
    Create a schedule

    NAME: Name of the App to run
    CRON: crontab representing the schedule interval
    PARAMETERS: key/value list of parameters to use when running the app on schedule
    [Project]: Project in which the app is present
    """
    params = process_cmd_param_vals(parameter)
    log.info(f"Adding schedule with parameters {params}")
    app_obj = api.LegacyApp.get(name, project=project)
    schedule_obj = api.Schedule.create(app_obj, cron, params)
    success_msg(f"Created schedule: {schedule_obj.id} ({schedule_obj.url})")


@schedule.command()
@click.option("--parameter", "-p", multiple=True)
@click.argument("id", required=True)
@click.argument("cron", required=False)
def update(id: str, cron: str, parameter: Tuple[str]):
    """
    Add a schedule

    ID: ID/URL of the Schedule
    CRON: crontab representing the schedule interval
    PARAMETERS: key/value list of parameters to use when running the app on schedule
    """

    params = process_cmd_param_vals(parameter)
    assert cron or parameter, "Must update either cron or parameters"

    log.info(f"Updating schedule with parameters {params}")

    schedule_obj = api.Schedule.by_id(id)
    schedule_obj.update(cron, params)
    success_msg(f"Updated schedule: {schedule_obj.id} ({schedule_obj.url})")


@schedule.command("list")
def schedule_list():
    """List all schedules"""
    print_table(api.Schedule.list(), "Schedules", showindex=False)


# @schedule.command()
# @click.argument("id", required=True)
# def get(id: str):
#     """Get variable value using variable name"""
#     res = api.Schedule.by_id(id)
#     print_table([res.dto], "Schedule")


@schedule.command()  # type: ignore[no-redef]
@click.argument("id", required=True)
def delete(id: str):
    """Delete a schedule by its id/url"""
    api.Schedule.by_id(id).delete()
    success_msg(f"Deleted schedule {id}")


def success_msg(msg: str):
    click.secho(msg, fg="green")


def failure_msg(msg: str, do_exit: bool = False):
    click.secho(msg, fg="red")
    if do_exit:
        ctx: click.Context = click.get_current_context(silent=True)
        if ctx:
            ctx.exit(2)
        else:
            exit(2)


#############################################################################

# Utilities for helping self-hosting apps


@app.group()
def generate():
    """Commands to generate misc files"""


@generate.command()
@click.option(
    "--python",
    default=hosting_utils.DEFAULT_PYTHON_VERSION,
    show_default=True,
)
@click.option(
    "--installer",
    default=hosting_utils.DEFAULT_INSTALLER,
    show_default=True,
    type=click.Choice(hosting_utils.SupportedInstallers.as_list(), case_sensitive=False),
)
@click.option(
    "--app-file",
    prompt="What file contains your App?",
    default=(lambda: next(hosting_utils.python_files(), None)),
    type=click.Path(exists=False, dir_okay=False),
    help="The App file you'd like to run (contains dp.serve_app(...))",
)
@click.option(
    "--output", "-o", default="Dockerfile", show_default=True, type=click.File(mode="w", lazy=True, atomic=True)
)
def dockerfile(python: str, installer: str, app_file: t.Union[str, Path, None], output: io.TextIOWrapper):
    """Generate a dockerfile to help with hosting Apps"""
    _installer = hosting_utils.SupportedInstallers[installer]

    if Path(output.name).exists() and output.name != "-":
        click.confirm(f"output file '{output.name}' exists. Overwrite?", default=False, abort=True)

    contents = hosting_utils.generate_dockerfile(python_version=python, installer=_installer, app_file=app_file)
    output.write(contents)


@generate.command()
@click.option(
    "--nb",
    "--app-file",
    "nb_path",
    prompt="What notebook would you like to convert?",
    default=(lambda: next(hosting_utils.python_files(extensions=[".ipynb"]), None)),
    type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
)
@click.option(
    "-o",
    "--out",
    "output",
    type=click.File(mode="w", lazy=True, atomic=True),
)
def script_from_nb(nb_path: t.Union[str, Path], output: t.Union[t.IO, None]):
    """Generate a script from a Jupyter Notebook

    Uses nbconvert to extract as a python script, with modifications
    to provide a more seamless experience.

    Outputs are not preserved.
    """
    from datapane.ipython.utils import extract_py_notebook

    nb_path = Path(nb_path)
    cwd = Path.cwd()
    if output is None:
        output_path = cwd / nb_path.with_suffix(".py").name
        output = click.open_file(str(output_path), mode="w", atomic=True)

    with click.open_file(str(nb_path), mode="r") as nb_file:
        py_contents = extract_py_notebook(nb_file)

    if output.name != "-":
        click.echo(f"Writing to file: {click.format_filename(output.name, shorten=True)}")
        if Path(output.name).exists():
            click.confirm("File exists. Overwrite?", default=False, abort=True)

    with output as f:
        f.write(py_contents)



"""Bottle plugins
TODO - do we just inline this where needed?
NOTE - basic plugin code vendored from Canister (https://github.com/dagnelies/canister - MIT License)

Canister is a simple plugin for bottle, providing:
- formatted logs
- url and form params unpacking
- sessions (server side) based on a `session_id` cookie
- authentication through basic auth or bearer token (OAuth2)
- CORS for cross-domain REST APIs

Changes Canister -> DPBottle
Removed
- Logging
- Global auth (both Basic and JWT)
  - instead a hardcoded "anonymous" user
- url / form param unpacking - unnecessary, prefer explicit usage
Added
- Typing and modern Python 3.7+ constructs
- General algorithm changes and improvements
"""
import math
import secrets
import threading
import time
import typing as t
from collections import UserDict
from contextlib import suppress
from functools import wraps

import datapane._vendor.bottle as bt
from datapane.client import log
from datapane.client.analytics import capture
from datapane.common.dp_types import SECS_1_HOUR, SECS_1_WEEK

from .runtime import GlobalState, SessionState

K = t.TypeVar("K")
V = t.TypeVar("V")

COOKIE_NAME = "dp_session_id"


class TimedDict(UserDict, t.Generic[K, V]):
    data: t.Dict[K, t.Tuple[float, V]]

    def __getitem__(self, key: K) -> V:
        # get the value and update the access time in the dict
        (_, val) = self.data[key]
        self.data[key] = (time.time(), val)
        return val

    def __setitem__(self, key: K, val: V):
        self.data[key] = (time.time(), val)

    def values(self):
        for _, val in self.data.values():
            yield val

    def items(self):
        for (k, (_, val)) in self.data.items():
            yield (k, val)

    def prune(self, age: float) -> int:
        # drop all values older than age and return the count
        now = time.time()
        initial_len = len(self.data)
        self.data = {k: (access_t, val) for (k, (access_t, val)) in self.data.items() if now - access_t < age}
        return initial_len - len(self.data)


class SessionCache(t.Generic[V]):
    """A thread safe session cache with a cleanup thread"""

    def __init__(self, timeout: int = SECS_1_HOUR):
        self._lock = threading.Lock()
        self._cache: TimedDict[str, V] = TimedDict()
        self.keep_pruning = True

        if timeout <= 0:
            log.warning("Sessions kept indefinitely! (session timeout is <= 0s)")
            return

        interval = int(math.sqrt(timeout))
        log.info(f"Session timeout is {timeout}s, checking for expired sessions every {interval}s")

        def prune() -> t.NoReturn:
            # NOTE - do we set this to falso on clear, or just keep always running when server is live?
            while self.keep_pruning:
                time.sleep(interval)
                with self._lock:
                    n = self._cache.prune(timeout)
                    log.debug(f"{n} expired sessions pruned")

        # daemon thread shuts down without cleanup - ok for now
        threading.Thread(name="SessionCleaner", target=prune, daemon=True).start()

    def __contains__(self, s_id: str) -> bool:
        return s_id in self._cache

    def get(self, s_id: str) -> t.Optional[V]:
        with self._lock:
            return self._cache.get(s_id, None)

    def set(self, s_id: str, data: V) -> None:
        assert s_id
        with self._lock:
            self._cache[s_id] = data

    def create(self, data: V) -> str:
        s_id: str = secrets.token_urlsafe(18)
        with self._lock:
            self._cache[s_id] = data
        return s_id

    def delete(self, s_id: str) -> None:
        with self._lock, suppress(KeyError):
            del self._cache[s_id]

    def clear(self) -> None:
        with self._lock:
            self._cache.clear()
            # self.keep_pruning = False


class DPBottlePlugin:
    """Specialised to the App Session object"""

    name = "dp_bottle"
    api = 2

    # attributes
    app: bt.Bottle
    sessions: SessionCache[SessionState]
    session_secret: str
    cors: t.Optional[str]
    g_s: GlobalState
    embed_mode: bool

    def __init__(self, g_s: GlobalState, embed_mode: bool):
        self.g_s = g_s
        self.embed_mode = embed_mode

    def setup(self, app: bt.Bottle):
        # if 'canister' not in app.config:
        #    raise Exception('Canister requires a configuration file. Please refer to the docs.')

        # log config
        log.debug("Initializing Bottle and custom plugin")
        log.debug(f"bottle version: {bt.__version__}")
        log.debug("------------------------------------------")
        for k, v in app.config.items():
            log.debug(f"{k:<30} = {v}")
        log.debug("------------------------------------------")

        app.log = log
        self.app = app

        timeout = int(app.config.get("dp_bottle.session_timeout", SECS_1_HOUR))
        self.sessions = SessionCache(timeout=timeout)
        self.session_secret = secrets.token_urlsafe(30)

        self.cors: str = app.config.get("dp_bottle.CORS", "").lower()
        if self.cors in ("", "false"):
            self.cors = None

    def _create_session(self, res: bt.LocalResponse) -> SessionState:
        """Create a new session and return in the response"""
        session = SessionState(user="anonymous")
        session.add_entry(self.g_s.main)

        s_id = session.s_id
        self.sessions.set(s_id, session)
        # NOTE - do we want a timelimit on cookie, or just tie to browser session?
        # if embed_mode is True (default: False), we disable samesite cookies so sessions work in iframes
        res.set_cookie(
            COOKIE_NAME,
            s_id,
            secret=self.session_secret,
            path="/",
            maxage=SECS_1_WEEK,
            secure=self.embed_mode,
            httponly=True,
            samesite="none" if self.embed_mode else "lax",
        )
        log.info(f"Session created: {s_id}")
        return session

    def apply(self, callback: t.Callable, route: bt.Route):
        @wraps(callback)
        def wrapper(*args, **kwargs):

            # grab the session threadlocal and attach to request
            # global session
            req = bt.request
            res = bt.response
            start = time.time()

            # thread name = <ip>-....
            threading.current_thread().name = f"{req.remote_addr}-..."
            log.info(f"{req.method} {req.url}")

            # session handling
            prev_s_id: t.Optional[str] = req.get_cookie(COOKIE_NAME, secret=self.session_secret)
            prev_session = self.sessions.get(prev_s_id)

            if route.rule == "/control/reset/" and route.method == "POST":
                # reset the session
                log.info(f"Resetting session for {prev_s_id=}")
                # remove the user session, if exists, and create a new one
                if prev_session is not None:
                    self.sessions.delete(prev_s_id)
                _ = self._create_session(res)
                res.status = 204
                return None
            elif prev_session is not None:
                session = t.cast(SessionState, prev_session)
                s_id = t.cast(str, prev_s_id)
                log.info(f"Existing session found for {s_id=}")
            elif route.rule == "/":  # and no prev session
                session = self._create_session(res)
                capture("App Session Created")
            else:
                # no valid previous session and calling a route other than "/"
                res.headers["Datapane-Error"] = "unknown-session"
                raise bt.HTTPError(400, "Unknown session", **res.headers)

            # thread name = <ip>-<session_id[0:6]>
            threading.current_thread().name = f"{req.remote_addr}-{session.s_id[0:6]}"

            # NOTE - user handling was here...
            # session.session.user = user = "anonymous"

            # store session directly in the threadlocal request
            _session_hash = hash(session)
            req.session = session

            # call the route dispatcher
            try:
                result = callback(*args, **kwargs)
            except OSError as e:
                # Catch OSError, as `socket` will raise OSError or TimeoutError depending on
                # where it is.
                #
                # occurs when 2 conditions are met:
                #   - the client has sent part of the body, but took too long to finish
                #   - we are currently trying to read the body
                #
                # This can be replicated using 'gnutls-cli' or 'telnet' to:
                # 1. connect to the server
                # 2. send a POST request with headers and the **start** of the body
                # 3. wait for the server to timeout the connection (~10 seconds from connection)
                #
                # In these cases, we don't need to provide a response

                # TimeoutError could happen in the handler.
                # Do a sanity check to make sure it's actually a socket error
                try:
                    _ = req.body
                except OSError:
                    # we were right, bail out
                    log.warning("dropping timed-out request")
                    return
                else:
                    raise e from e

            # update the session, if needed
            if _session_hash != hash(session):
                self.sessions.set(session.s_id, session)

            # Cors handling
            if self.cors:
                res.headers["Access-Control-Allow-Origin"] = self.cors

            # request timing
            elapsed = time.time() - start
            if elapsed > 1:
                log.warning(f"Response: {res.status_code} ({1000*elapsed:.2f}ms - Slow!)")
            else:
                log.info(f"Response: {res.status_code} ({1000 * elapsed:.2f}ms)")
            return result

        return wrapper

    def close(self):
        log.debug("Closing DPBottle Plugin")
        self.sessions.clear()
        self.g_s.clear()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


import pytest

import pyarrow as pa
from pyarrow import compute as pc

# UDFs are all tested with a dataset scan
pytestmark = pytest.mark.dataset


try:
    import pyarrow.dataset as ds
except ImportError:
    ds = None


def mock_udf_context(batch_length=10):
    from pyarrow._compute import _get_scalar_udf_context
    return _get_scalar_udf_context(pa.default_memory_pool(), batch_length)


class MyError(RuntimeError):
    pass


@pytest.fixture(scope="session")
def unary_func_fixture():
    """
    Register a unary scalar function.
    """
    def unary_function(ctx, x):
        return pc.call_function("add", [x, 1],
                                memory_pool=ctx.memory_pool)
    func_name = "y=x+1"
    unary_doc = {"summary": "add function",
                 "description": "test add function"}
    pc.register_scalar_function(unary_function,
                                func_name,
                                unary_doc,
                                {"array": pa.int64()},
                                pa.int64())
    return unary_function, func_name


@pytest.fixture(scope="session")
def binary_func_fixture():
    """
    Register a binary scalar function.
    """
    def binary_function(ctx, m, x):
        return pc.call_function("multiply", [m, x],
                                memory_pool=ctx.memory_pool)
    func_name = "y=mx"
    binary_doc = {"summary": "y=mx",
                  "description": "find y from y = mx"}
    pc.register_scalar_function(binary_function,
                                func_name,
                                binary_doc,
                                {"m": pa.int64(),
                                 "x": pa.int64(),
                                 },
                                pa.int64())
    return binary_function, func_name


@pytest.fixture(scope="session")
def ternary_func_fixture():
    """
    Register a ternary scalar function.
    """
    def ternary_function(ctx, m, x, c):
        mx = pc.call_function("multiply", [m, x],
                              memory_pool=ctx.memory_pool)
        return pc.call_function("add", [mx, c],
                                memory_pool=ctx.memory_pool)
    ternary_doc = {"summary": "y=mx+c",
                   "description": "find y from y = mx + c"}
    func_name = "y=mx+c"
    pc.register_scalar_function(ternary_function,
                                func_name,
                                ternary_doc,
                                {
                                    "array1": pa.int64(),
                                    "array2": pa.int64(),
                                    "array3": pa.int64(),
                                },
                                pa.int64())
    return ternary_function, func_name


@pytest.fixture(scope="session")
def varargs_func_fixture():
    """
    Register a varargs scalar function with at least two arguments.
    """
    def varargs_function(ctx, first, *values):
        acc = first
        for val in values:
            acc = pc.call_function("add", [acc, val],
                                   memory_pool=ctx.memory_pool)
        return acc
    func_name = "z=ax+by+c"
    varargs_doc = {"summary": "z=ax+by+c",
                   "description": "find z from z = ax + by + c"
                   }
    pc.register_scalar_function(varargs_function,
                                func_name,
                                varargs_doc,
                                {
                                    "array1": pa.int64(),
                                    "array2": pa.int64(),
                                },
                                pa.int64())
    return varargs_function, func_name


@pytest.fixture(scope="session")
def nullary_func_fixture():
    """
    Register a nullary scalar function.
    """
    def nullary_func(context):
        return pa.array([42] * context.batch_length, type=pa.int64(),
                        memory_pool=context.memory_pool)

    func_doc = {
        "summary": "random function",
        "description": "generates a random value"
    }
    func_name = "test_nullary_func"
    pc.register_scalar_function(nullary_func,
                                func_name,
                                func_doc,
                                {},
                                pa.int64())

    return nullary_func, func_name


@pytest.fixture(scope="session")
def wrong_output_type_func_fixture():
    """
    Register a scalar function which returns something that is neither
    a Arrow scalar or array.
    """
    def wrong_output_type(ctx):
        return 42

    func_name = "test_wrong_output_type"
    in_types = {}
    out_type = pa.int64()
    doc = {
        "summary": "return wrong output type",
        "description": ""
    }
    pc.register_scalar_function(wrong_output_type, func_name, doc,
                                in_types, out_type)
    return wrong_output_type, func_name


@pytest.fixture(scope="session")
def wrong_output_datatype_func_fixture():
    """
    Register a scalar function whose actual output DataType doesn't
    match the declared output DataType.
    """
    def wrong_output_datatype(ctx, array):
        return pc.call_function("add", [array, 1])
    func_name = "test_wrong_output_datatype"
    in_types = {"array": pa.int64()}
    # The actual output DataType will be int64.
    out_type = pa.int16()
    doc = {
        "summary": "return wrong output datatype",
        "description": ""
    }
    pc.register_scalar_function(wrong_output_datatype, func_name, doc,
                                in_types, out_type)
    return wrong_output_datatype, func_name


@pytest.fixture(scope="session")
def wrong_signature_func_fixture():
    """
    Register a scalar function with the wrong signature.
    """
    # Missing the context argument
    def wrong_signature():
        return pa.scalar(1, type=pa.int64())

    func_name = "test_wrong_signature"
    in_types = {}
    out_type = pa.int64()
    doc = {
        "summary": "UDF with wrong signature",
        "description": ""
    }
    pc.register_scalar_function(wrong_signature, func_name, doc,
                                in_types, out_type)
    return wrong_signature, func_name


@pytest.fixture(scope="session")
def raising_func_fixture():
    """
    Register a scalar function which raises a custom exception.
    """
    def raising_func(ctx):
        raise MyError("error raised by scalar UDF")
    func_name = "test_raise"
    doc = {
        "summary": "raising function",
        "description": ""
    }
    pc.register_scalar_function(raising_func, func_name, doc,
                                {}, pa.int64())
    return raising_func, func_name


def check_scalar_function(func_fixture,
                          inputs, *,
                          run_in_dataset=True,
                          batch_length=None):
    function, name = func_fixture
    if batch_length is None:
        all_scalar = True
        for arg in inputs:
            if isinstance(arg, pa.Array):
                all_scalar = False
                batch_length = len(arg)
        if all_scalar:
            batch_length = 1

    expected_output = function(mock_udf_context(batch_length), *inputs)
    func = pc.get_function(name)
    assert func.name == name

    result = pc.call_function(name, inputs, length=batch_length)
    assert result == expected_output
    # At the moment there is an issue when handling nullary functions.
    # See: ARROW-15286 and ARROW-16290.
    if run_in_dataset:
        field_names = [f'field{index}' for index, in_arr in inputs]
        table = pa.Table.from_arrays(inputs, field_names)
        dataset = ds.dataset(table)
        func_args = [ds.field(field_name) for field_name in field_names]
        result_table = dataset.to_table(
            columns={'result': ds.field('')._call(name, func_args)})
        assert result_table.column(0).chunks[0] == expected_output


def test_scalar_udf_array_unary(unary_func_fixture):
    check_scalar_function(unary_func_fixture,
                          [
                              pa.array([10, 20], pa.int64())
                          ]
                          )


def test_scalar_udf_array_binary(binary_func_fixture):
    check_scalar_function(binary_func_fixture,
                          [
                              pa.array([10, 20], pa.int64()),
                              pa.array([2, 4], pa.int64())
                          ]
                          )


def test_scalar_udf_array_ternary(ternary_func_fixture):
    check_scalar_function(ternary_func_fixture,
                          [
                              pa.array([10, 20], pa.int64()),
                              pa.array([2, 4], pa.int64()),
                              pa.array([5, 10], pa.int64())
                          ]
                          )


def test_scalar_udf_array_varargs(varargs_func_fixture):
    check_scalar_function(varargs_func_fixture,
                          [
                              pa.array([2, 3], pa.int64()),
                              pa.array([10, 20], pa.int64()),
                              pa.array([3, 7], pa.int64()),
                              pa.array([20, 30], pa.int64()),
                              pa.array([5, 10], pa.int64())
                          ]
                          )


def test_registration_errors():
    # validate function name
    doc = {
        "summary": "test udf input",
        "description": "parameters are validated"
    }
    in_types = {"scalar": pa.int64()}
    out_type = pa.int64()

    def test_reg_function(context):
        return pa.array([10])

    with pytest.raises(TypeError):
        pc.register_scalar_function(test_reg_function,
                                    None, doc, in_types,
                                    out_type)

    # validate function
    with pytest.raises(TypeError, match="func must be a callable"):
        pc.register_scalar_function(None, "test_none_function", doc, in_types,
                                    out_type)

    # validate output type
    expected_expr = "DataType expected, got <class 'NoneType'>"
    with pytest.raises(TypeError, match=expected_expr):
        pc.register_scalar_function(test_reg_function,
                                    "test_output_function", doc, in_types,
                                    None)

    # validate input type
    expected_expr = "in_types must be a dictionary of DataType"
    with pytest.raises(TypeError, match=expected_expr):
        pc.register_scalar_function(test_reg_function,
                                    "test_input_function", doc, None,
                                    out_type)

    # register an already registered function
    # first registration
    pc.register_scalar_function(test_reg_function,
                                "test_reg_function", doc, {},
                                out_type)
    # second registration
    expected_expr = "Already have a function registered with name:" \
        + " test_reg_function"
    with pytest.raises(KeyError, match=expected_expr):
        pc.register_scalar_function(test_reg_function,
                                    "test_reg_function", doc, {},
                                    out_type)


def test_varargs_function_validation(varargs_func_fixture):
    _, func_name = varargs_func_fixture

    error_msg = r"VarArgs function 'z=ax\+by\+c' needs at least 2 arguments"

    with pytest.raises(ValueError, match=error_msg):
        pc.call_function(func_name, [42])


def test_function_doc_validation():
    # validate arity
    in_types = {"scalar": pa.int64()}
    out_type = pa.int64()

    # doc with no summary
    func_doc = {
        "description": "desc"
    }

    def add_const(ctx, scalar):
        return pc.call_function("add", [scalar, 1])

    with pytest.raises(ValueError,
                       match="Function doc must contain a summary"):
        pc.register_scalar_function(add_const, "test_no_summary",
                                    func_doc, in_types,
                                    out_type)

    # doc with no decription
    func_doc = {
        "summary": "test summary"
    }

    with pytest.raises(ValueError,
                       match="Function doc must contain a description"):
        pc.register_scalar_function(add_const, "test_no_desc",
                                    func_doc, in_types,
                                    out_type)


def test_nullary_function(nullary_func_fixture):
    # XXX the Python compute layer API doesn't let us override batch_length,
    # so only test with the default value of 1.
    check_scalar_function(nullary_func_fixture, [], run_in_dataset=False,
                          batch_length=1)


def test_wrong_output_type(wrong_output_type_func_fixture):
    _, func_name = wrong_output_type_func_fixture

    with pytest.raises(TypeError,
                       match="Unexpected output type: int"):
        pc.call_function(func_name, [], length=1)


def test_wrong_output_datatype(wrong_output_datatype_func_fixture):
    _, func_name = wrong_output_datatype_func_fixture

    expected_expr = ("Expected output datatype int16, "
                     "but function returned datatype int64")

    with pytest.raises(TypeError, match=expected_expr):
        pc.call_function(func_name, [pa.array([20, 30])])


def test_wrong_signature(wrong_signature_func_fixture):
    _, func_name = wrong_signature_func_fixture

    expected_expr = (r"wrong_signature\(\) takes 0 positional arguments "
                     "but 1 was given")

    with pytest.raises(TypeError, match=expected_expr):
        pc.call_function(func_name, [], length=1)


def test_wrong_datatype_declaration():
    def identity(ctx, val):
        return val

    func_name = "test_wrong_datatype_declaration"
    in_types = {"array": pa.int64()}
    out_type = {}
    doc = {
        "summary": "test output value",
        "description": "test output"
    }
    with pytest.raises(TypeError,
                       match="DataType expected, got <class 'dict'>"):
        pc.register_scalar_function(identity, func_name,
                                    doc, in_types, out_type)


def test_wrong_input_type_declaration():
    def identity(ctx, val):
        return val

    func_name = "test_wrong_input_type_declaration"
    in_types = {"array": None}
    out_type = pa.int64()
    doc = {
        "summary": "test invalid input type",
        "description": "invalid input function"
    }
    with pytest.raises(TypeError,
                       match="DataType expected, got <class 'NoneType'>"):
        pc.register_scalar_function(identity, func_name, doc,
                                    in_types, out_type)


def test_udf_context(unary_func_fixture):
    # Check the memory_pool argument is properly propagated
    proxy_pool = pa.proxy_memory_pool(pa.default_memory_pool())
    _, func_name = unary_func_fixture

    res = pc.call_function(func_name,
                           [pa.array([1] * 1000, type=pa.int64())],
                           memory_pool=proxy_pool)
    assert res == pa.array([2] * 1000, type=pa.int64())
    assert proxy_pool.bytes_allocated() == 1000 * 8
    # Destroying Python array should destroy underlying C++ memory
    res = None
    assert proxy_pool.bytes_allocated() == 0


def test_raising_func(raising_func_fixture):
    _, func_name = raising_func_fixture
    with pytest.raises(MyError, match="error raised by scalar UDF"):
        pc.call_function(func_name, [], length=1)


def test_scalar_input(unary_func_fixture):
    function, func_name = unary_func_fixture
    res = pc.call_function(func_name, [pa.scalar(10)])
    assert res == pa.scalar(11)


def test_input_lifetime(unary_func_fixture):
    function, func_name = unary_func_fixture

    proxy_pool = pa.proxy_memory_pool(pa.default_memory_pool())
    assert proxy_pool.bytes_allocated() == 0

    v = pa.array([1] * 1000, type=pa.int64(), memory_pool=proxy_pool)
    assert proxy_pool.bytes_allocated() == 1000 * 8
    pc.call_function(func_name, [v])
    assert proxy_pool.bytes_allocated() == 1000 * 8
    # Calling a UDF should not have kept `v` alive longer than required
    v = None
    assert proxy_pool.bytes_allocated() == 0



import dataclasses as dc
import json
from typing import List, Optional, TextIO

from datapane.common import Hash


@dc.dataclass
class RunResult:
    # TODO - this should be encrypted using django.sign - however is a hashid for now...
    report_id: Optional[str] = None
    script_result: Optional[str] = None
    output: Optional[str] = None
    cacheable: bool = False
    # used for GC - tho could get by xpath
    cas_refs: List[Hash] = dc.field(default_factory=list)
    asset_ids: List[int] = dc.field(default_factory=list)

    def to_json(self, stream: TextIO):
        json.dump(dc.asdict(self), stream)


@dc.dataclass
class ErrorResult:
    error: str
    error_detail: str = ""
    output: Optional[str] = None
    debug: Optional[str] = None

    def to_json(self, stream: TextIO):
        json.dump(dc.asdict(self), stream)




import pytest

from itertools import count, dropwhile, chain

from glom import Iter
from glom import glom, SKIP, STOP, T, Call, Spec, Glommer, Check, SKIP


RANGE_5 = list(range(5))


def test_iter():
    assert list(glom(['1', '2', '3'], Iter(int))) == [1, 2, 3]
    cnt = count()
    cnt_1 = glom(cnt, Iter(lambda t: t + 1))
    assert (next(cnt_1), next(cnt_1)) == (1, 2)
    assert next(cnt) == 2

    assert list(glom(['1', '2', '3'], (Iter(int), enumerate))) == [(0, 1), (1, 2), (2, 3)]

    assert list(glom([1, SKIP, 2], Iter())) == [1, 2]
    assert list(glom([1, STOP, 2], Iter())) == [1]

    with pytest.raises(TypeError):
        Iter(nonexistent_kwarg=True)


def test_filter():
    is_odd = lambda x: x % 2
    odd_spec = Iter().filter(is_odd)
    out = glom(RANGE_5, odd_spec)
    assert list(out) == [1, 3]

    # let's just make sure we're actually streaming just in case
    counter = count()
    out = glom(counter, odd_spec)
    assert next(out) == 1
    assert next(out) == 3
    assert next(counter) == 4
    assert next(counter) == 5
    assert next(out) == 7

    bools = [True, False, False, True, False]
    spec = Iter().filter().all()
    out = glom(bools, spec)
    assert out == [True, True]

    imags = [0j, 1j, 2, 2j, 3j]
    spec = Iter().filter(Check(T.imag.real, type=float, one_of=(0, 2), default=SKIP)).all()
    out = glom(imags, spec)
    assert out == [0j, 2j]

    assert repr(Iter().filter(T.a.b)) == 'Iter().filter(T.a.b)'
    assert repr(Iter(list).filter(sum)) == 'Iter(list).filter(sum)'


def test_map():
    spec = Iter().map(lambda x: x * 2)
    out = glom(RANGE_5, spec)
    assert list(out) == [0, 2, 4, 6, 8]
    assert repr(Iter().map(T.a.b)) == 'Iter().map(T.a.b)'


def test_split_flatten():
    falsey_stream = [1, None, None, 2, 3, None, 4]
    spec = Iter().split()
    out = glom(falsey_stream, spec)
    assert list(out) == [[1], [2, 3], [4]]

    spec = Iter().split().flatten()
    out = glom(falsey_stream, spec)
    assert list(out) == [1, 2, 3, 4]

    assert repr(Iter().split(sep=None, maxsplit=2)) == 'Iter().split(sep=None, maxsplit=2)'
    assert repr(Iter(T.a.b[1]).flatten()) == 'Iter(T.a.b[1]).flatten()'


def test_chunked():
    int_list = list(range(9))

    spec = Iter().chunked(3)
    out = glom(int_list, spec)
    assert list(out) == [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

    spec = Iter().chunked(3).map(sum)
    out = glom(int_list, spec)
    assert list(out) == [3, 12, 21]


def test_windowed():
    int_list = list(range(5))

    spec = Iter().windowed(3)
    out = glom(int_list, spec)
    assert list(out) == [(0, 1, 2), (1, 2, 3), (2, 3, 4)]
    assert repr(spec) == 'Iter().windowed(3)'

    spec = spec.filter(lambda x: bool(x[0] % 2)).map(sum)
    out = glom(int_list, spec)
    assert next(out) == 6

    out = glom(range(10), spec)
    assert list(out) == [6, 12, 18, 24]


def test_unique():
    int_list = list(range(10))

    spec = Iter().unique()
    out = glom(int_list, spec)
    assert list(out) == int_list

    spec = Iter(lambda x: x % 4).unique()
    out = glom(int_list, spec)
    assert list(out) == int_list[:4]
    assert repr(Iter().unique(T.a)) == 'Iter().unique(T.a)'


def test_slice():
    cnt = count()

    spec = Iter().slice(3)
    out = glom(cnt, spec)

    assert list(out) == [0, 1, 2]
    assert next(cnt) == 3

    out = glom(range(10), Iter().slice(1, 5))
    assert list(out) == [1, 2, 3, 4]

    out = glom(range(10), Iter().slice(1, 6, 2))
    assert list(out) == [1, 3, 5]
    assert repr(Iter().slice(1, 6, 2)) == 'Iter().slice(1, 6, 2)'

    out = glom(range(10), Iter().limit(3))
    assert list(out) == [0, 1, 2]
    assert repr(Iter().limit(3)) == 'Iter().limit(3)'

    out = glom(range(5), Iter().limit(10))
    assert list(out) == [0, 1, 2, 3, 4]

    # test broken args
    with pytest.raises(TypeError):
        Iter().slice(1, 2, 3, 4)


def test_while():
    cnt = count()
    out = glom(cnt, Iter().takewhile(lambda x: x < 3))
    assert list(out) == [0, 1, 2]
    assert next(cnt) == 4
    assert repr(Iter().takewhile(T.a) == 'Iter().takewhile(T.a)')

    range_iter = iter(range(7))
    out = glom(range_iter, Iter().dropwhile(lambda x: x < 3 or x > 5))
    assert list(out) == [3, 4, 5, 6]  # 6 still here despite the x>5 above

    out = glom(range(10), Iter().dropwhile(lambda x: x >= 0).limit(10))
    assert list(out) == []

    out = glom(range(8), Iter().dropwhile((T.bit_length(), lambda x: x < 3)))
    assert list(out) == [4, 5, 6, 7]
    assert repr(Iter().dropwhile(T.a) == 'Iter().dropwhile(T.a)')


def test_iter_composition():
    int_list = list(range(10))
    out = glom(int_list, (Iter(), Iter(), list))
    assert out == int_list

    out = glom([int_list] * 3, Iter(Iter(lambda x: x % 4)).flatten().unique())
    assert list(out) == [0, 1, 2, 3]


def test_faulty_iterate():
    glommer = Glommer()

    def bad_iter(obj):
        raise RuntimeError('oops')

    glommer.register(str, iterate=bad_iter)

    with pytest.raises(TypeError):
        glommer.glom('abc', (Iter(), list))


def test_first():
    spec = Iter().first(T.imag)

    target = iter([1, 2, 3j, 4])
    out = glom(target, spec)
    assert out == 3j
    assert next(target) == 4
    assert repr(spec) == '(Iter(), First(T.imag))'

    spec = Iter().first(T.imag, default=0)
    target = iter([1, 2, 4])
    out = glom(target, spec)
    assert out == 0
    assert repr(spec) == '(Iter(), First(T.imag, default=0))'


def test_all():
    int_iter = iter(range(10))

    out = glom(int_iter, Iter().all())
    assert out == list(range(10))
    assert next(int_iter, None) is None
    assert repr(Iter().all()) == 'Pipe(Iter(), list)'



import logging
from typing import Dict, Optional

from aws_lambda_powertools.event_handler.api_gateway import Response
from aws_lambda_powertools.event_handler.exceptions import BadRequestError, InternalServerError
from aws_lambda_powertools.event_handler.middlewares import BaseMiddlewareHandler, NextMiddleware
from aws_lambda_powertools.event_handler.types import EventHandlerInstance
from aws_lambda_powertools.utilities.validation import validate
from aws_lambda_powertools.utilities.validation.exceptions import InvalidSchemaFormatError, SchemaValidationError

logger = logging.getLogger(__name__)


class SchemaValidationMiddleware(BaseMiddlewareHandler):
    """Middleware to validate API request and response against JSON Schema using the [Validation utility](https://docs.powertools.aws.dev/lambda/python/latest/utilities/validation/).

    Examples
    --------
    **Validating incoming event**

    ```python
    import requests

    from aws_lambda_powertools import Logger
    from aws_lambda_powertools.event_handler import APIGatewayRestResolver, Response
    from aws_lambda_powertools.event_handler.middlewares import BaseMiddlewareHandler, NextMiddleware
    from aws_lambda_powertools.event_handler.middlewares.schema_validation import SchemaValidationMiddleware

    app = APIGatewayRestResolver()
    logger = Logger()
    json_schema_validation = SchemaValidationMiddleware(inbound_schema=INCOMING_JSON_SCHEMA)


    @app.get("/todos", middlewares=[json_schema_validation])
    def get_todos():
        todos: requests.Response = requests.get("https://jsonplaceholder.typicode.com/todos")
        todos.raise_for_status()

        # for brevity, we'll limit to the first 10 only
        return {"todos": todos.json()[:10]}


    @logger.inject_lambda_context
    def lambda_handler(event, context):
        return app.resolve(event, context)
    ```
    """

    def __init__(
        self,
        inbound_schema: Dict,
        inbound_formats: Optional[Dict] = None,
        outbound_schema: Optional[Dict] = None,
        outbound_formats: Optional[Dict] = None,
    ):
        """See [Validation utility](https://docs.powertools.aws.dev/lambda/python/latest/utilities/validation/) docs for examples on all parameters.

        Parameters
        ----------
        inbound_schema : Dict
            JSON Schema to validate incoming event
        inbound_formats : Optional[Dict], optional
            Custom formats containing a key (e.g. int64) and a value expressed as regex or callback returning bool, by default None
            JSON Schema to validate outbound event, by default None
        outbound_formats : Optional[Dict], optional
            Custom formats containing a key (e.g. int64) and a value expressed as regex or callback returning bool, by default None
        """  # noqa: E501
        super().__init__()
        self.inbound_schema = inbound_schema
        self.inbound_formats = inbound_formats
        self.outbound_schema = outbound_schema
        self.outbound_formats = outbound_formats

    def bad_response(self, error: SchemaValidationError) -> Response:
        message: str = f"Bad Response: {error.message}"
        logger.debug(message)
        raise BadRequestError(message)

    def bad_request(self, error: SchemaValidationError) -> Response:
        message: str = f"Bad Request: {error.message}"
        logger.debug(message)
        raise BadRequestError(message)

    def bad_config(self, error: InvalidSchemaFormatError) -> Response:
        logger.debug(f"Invalid Schema Format: {error}")
        raise InternalServerError("Internal Server Error")

    def handler(self, app: EventHandlerInstance, next_middleware: NextMiddleware) -> Response:
        """Validates incoming JSON payload (body) against JSON Schema provided.

        Parameters
        ----------
        app : EventHandlerInstance
            An instance of an Event Handler
        next_middleware : NextMiddleware
            Callable to get response from the next middleware or route handler in the chain

        Returns
        -------
        Response
            It can return three types of response objects

            - Original response: Propagates HTTP response returned from the next middleware if validation succeeds
            - HTTP 400: Payload or response failed JSON Schema validation
            - HTTP 500: JSON Schema provided has incorrect format
        """
        try:
            validate(event=app.current_event.json_body, schema=self.inbound_schema, formats=self.inbound_formats)
        except SchemaValidationError as error:
            return self.bad_request(error)
        except InvalidSchemaFormatError as error:
            return self.bad_config(error)

        result = next_middleware(app)

        if self.outbound_formats is not None:
            try:
                validate(event=result.body, schema=self.inbound_schema, formats=self.inbound_formats)
            except SchemaValidationError as error:
                return self.bad_response(error)
            except InvalidSchemaFormatError as error:
                return self.bad_config(error)

        return result



# Copyright 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Data Commons Python API Core.

Provides primitive operations for working with collections of nodes. For a
collection of nodes identified by their dcids, this submodule implements the
following:

- Getting all property labels
- Getting all property values
- Getting all triples
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import defaultdict

import datacommons.utils as utils

# ----------------------------- WRAPPER FUNCTIONS -----------------------------


def get_property_labels(dcids, out=True):
  """ Returns the labels of properties defined for the given :code:`dcids`.

  Args:
    dcids (:obj:`iterable` of :obj:`str`): A list of nodes identified by their
      dcids.
    out (:obj:`bool`, optional): Whether or not the property points away from
      the given list of nodes.

  Returns:
    A :obj:`dict` mapping dcids to lists of property labels. If `out` is `True`,
    then property labels correspond to edges directed away from given nodes.
    Otherwise, they correspond to edges directed towards the given nodes.

  Raises:
    ValueError: If the payload returned by the Data Commons REST API is
      malformed.

  Examples:
    To get all outgoing property labels for
    `California <https://browser.datacommons.org/kg?dcid=geoId/06>`_ and
    `Colorado <https://browser.datacommons.org/kg?dcid=geoId/08>`_, we can write
    the following.

    >>> get_property_labels(['geoId/06', 'geoId/08'])
    {
      "geoId/06": [
        "containedInPlace",
        "geoId",
        "kmlCoordinates",
        "name",
        "provenance",
        "typeOf"
      ],
      "geoId/08",: [
        "containedInPlace",
        "geoId",
        "kmlCoordinates",
        "name",
        "provenance",
        "typeOf"
      ]
    }

    We can also get incoming property labels by setting `out=False`.

    >>> get_property_labels(['geoId/06', 'geoId/08'], out=False)
    {
      "geoId/06": [
        "addressRegion",
        "containedInPlace",
        "location",
        "overlapsWith"
      ],
      "geoId/08",: [
        "addressRegion",
        "containedInPlace",
        "location",
        "overlapsWith"
      ]
    }
  """
  # Generate the GetProperty query and send the request
  dcids = filter(lambda v: v==v, dcids)  # Filter out NaN values
  dcids = list(dcids)
  url = utils._API_ROOT + utils._API_ENDPOINTS['get_property_labels']
  payload = utils._send_request(url, req_json={'dcids': dcids})

  # Return the results based on the orientation
  results = {}
  for dcid in dcids:
    if out:
      results[dcid] = payload[dcid]['outLabels']
    else:
      results[dcid] = payload[dcid]['inLabels']
  return results


def get_property_values(dcids,
                        prop,
                        out=True,
                        value_type=None,
                        limit=utils._MAX_LIMIT):
  """ Returns property values of given :code:`dcids` along the given property.

  Args:
    dcids (:obj:`iterable` of :obj:`str`): dcids to get property values for.
    prop (:obj:`str`): The property to get property values for.
    out (:obj:`bool`, optional): A flag that indicates the property is directed
      away from the given nodes when set to true.
    value_type (:obj:`str`, optional): A type to filter returned property values
      by.
    limit (:obj:`int`, optional): The maximum number of property values returned
      aggregated over all given nodes.

  Returns:
    Returned property values are formatted as a :obj:`dict` from a given dcid
    to a list of its property values.

  Raises:
    ValueError: If the payload returned by the Data Commons REST API is
      malformed.

  Examples:
    We would like to get the `name` of a list of states specified by their dcid:
    `geoId/06 <https://browser.datacommons.org/kg?dcid=geoId/06>`_,
    `geoId/21 <https://browser.datacommons.org/kg?dcid=geoId/21>`_, and
    `geoId/24 <https://browser.datacommons.org/kg?dcid=geoId/24>`_

    First, let's try specifying the :code:`dcids` as a :obj:`list` of
    :obj:`str`.

    >>> get_property_values(["geoId/06", "geoId/21", "geoId/24"], "name")
    {
      "geoId/06": ["California"],
      "geoId/21": ["Kentucky"],
      "geoId/24": ["Maryland"],
    }
  """
  # Convert the dcids field and format the request to GetPropertyValue
  dcids = filter(lambda v: v==v, dcids)  # Filter out NaN values
  dcids = list(dcids)
  if out:
    direction = 'out'
  else:
    direction = 'in'

  req_json = {
    'dcids': dcids,
    'property': prop,
    'limit': limit,
    'direction': direction
  }
  if value_type:
    req_json['value_type'] = value_type

  # Send the request
  url = utils._API_ROOT + utils._API_ENDPOINTS['get_property_values']
  payload = utils._send_request(url, req_json=req_json)

  # Create the result format for when dcids is provided as a list.
  unique_results = defaultdict(set)
  for dcid in dcids:
    # Get the list of nodes based on the direction given.
    nodes = []
    if out:
      if dcid in payload and 'out' in payload[dcid]:
        nodes = payload[dcid]['out']
    else:
       if dcid in payload and 'in' in payload[dcid]:
          nodes = payload[dcid]['in']

    # Add nodes to unique_results if it is not empty
    for node in nodes:
      if 'dcid' in node:
        unique_results[dcid].add(node['dcid'])
      elif 'value' in node:
        unique_results[dcid].add(node['value'])

  # Make sure each dcid is in the results dict, and convert all sets to lists.
  results = {dcid: sorted(list(unique_results[dcid])) for dcid in dcids}

  return results


def get_triples(dcids, limit=utils._MAX_LIMIT):
  """ Returns all triples associated with the given :code:`dcids`.

  A knowledge graph can be described as a collection of `triples` which are
  3-tuples that take the form `(s, p, o)`. Here `s` and `o` are nodes in the
  graph called the *subject* and *object* respectively while `p` is the property
  label of a directed edge from `s` to `o` (sometimes also called the
  *predicate*).

  Args:
    dcids (:obj:`iterable` of :obj:`str`): A list of dcids to get triples for.
    limit (:obj:`int`, optional): The maximum total number of triples to get.

  Returns:
    A :obj:`dict` mapping dcids to a :obj:`list` of triples `(s, p, o)` where
    `s`, `p`, and `o` are instances of :obj:`str` and either the subject
    or object is the mapped dcid.

  Raises:
    ValueError: If the payload returned by the Data Commons REST API is
      malformed.

  Examples:
    We would like to get five triples associated with
    `California <https://browser.datacommons.org/kg?dcid=geoId/06>`_

    >>> get_triples(["geoId/06"], limit=5)
    {
      "geoId/06": [
        ("geoId/06", "name", "California"),
        ("geoId/06", "typeOf", "State"),
        ("geoId/06", "geoId", "06"),
        ("geoId/0687056", "containedInPlace", "geoId/06"),
        ("geoId/0686440", "containedInPlace", "geoId/06")
      ]
    }
  """
  # Generate the GetTriple query and send the request.
  dcids = filter(lambda v: v==v, dcids)  # Filter out NaN values
  dcids = list(dcids)
  url = utils._API_ROOT + utils._API_ENDPOINTS['get_triples']
  payload = utils._send_request(url, req_json={'dcids': dcids, 'limit': limit})

  # Create a map from dcid to list of triples.
  results = defaultdict(list)
  for dcid in dcids:
    # Make sure each dcid is mapped to an empty list.
    results[dcid]

    # Add triples as appropriate
    for t in payload[dcid]:
      if 'objectId' in t:
        results[dcid].append(
          (t['subjectId'], t['predicate'], t['objectId']))
      elif 'objectValue' in t:
        results[dcid].append(
          (t['subjectId'], t['predicate'], t['objectValue']))
  return dict(results)



from typing import Any, Dict, Optional, overload

from aws_lambda_powertools.shared.headers_serializer import (
    BaseHeadersSerializer,
    HttpApiHeadersSerializer,
)
from aws_lambda_powertools.utilities.data_classes.common import BaseProxyEvent, DictWrapper
from aws_lambda_powertools.utilities.data_classes.shared_functions import (
    base64_decode,
    get_header_value,
    get_query_string_value,
)


class VPCLatticeEventBase(BaseProxyEvent):
    @property
    def body(self) -> str:
        """The VPC Lattice body."""
        return self["body"]

    @property
    def json_body(self) -> Any:
        """Parses the submitted body as json"""
        if self._json_data is None:
            self._json_data = self._json_deserializer(self.decoded_body)
        return self._json_data

    @property
    def headers(self) -> Dict[str, str]:
        """The VPC Lattice event headers."""
        return self["headers"]

    @property
    def decoded_body(self) -> str:
        """Dynamically base64 decode body as a str"""
        body: str = self["body"]
        if self.is_base64_encoded:
            return base64_decode(body)
        return body

    @property
    def method(self) -> str:
        """The VPC Lattice method used. Valid values include: DELETE, GET, HEAD, OPTIONS, PATCH, POST, and PUT."""
        return self["method"]

    @property
    def http_method(self) -> str:
        """The HTTP method used. Valid values include: DELETE, GET, HEAD, OPTIONS, PATCH, POST, and PUT."""
        return self["method"]

    def get_query_string_value(self, name: str, default_value: Optional[str] = None) -> Optional[str]:
        """Get query string value by name

        Parameters
        ----------
        name: str
            Query string parameter name
        default_value: str, optional
            Default value if no value was found by name
        Returns
        -------
        str, optional
            Query string parameter value
        """
        return get_query_string_value(
            query_string_parameters=self.query_string_parameters,
            name=name,
            default_value=default_value,
        )

    @overload
    def get_header_value(
        self,
        name: str,
        default_value: str,
        case_sensitive: Optional[bool] = False,
    ) -> str:
        ...

    @overload
    def get_header_value(
        self,
        name: str,
        default_value: Optional[str] = None,
        case_sensitive: Optional[bool] = False,
    ) -> Optional[str]:
        ...

    def get_header_value(
        self,
        name: str,
        default_value: Optional[str] = None,
        case_sensitive: Optional[bool] = False,
    ) -> Optional[str]:
        """Get header value by name

        Parameters
        ----------
        name: str
            Header name
        default_value: str, optional
            Default value if no value was found by name
        case_sensitive: bool
            Whether to use a case-sensitive look up
        Returns
        -------
        str, optional
            Header value
        """
        return get_header_value(
            headers=self.headers,
            name=name,
            default_value=default_value,
            case_sensitive=case_sensitive,
        )

    def header_serializer(self) -> BaseHeadersSerializer:
        # When using the VPC Lattice integration, we have multiple HTTP Headers.
        return HttpApiHeadersSerializer()


class VPCLatticeEvent(VPCLatticeEventBase):
    @property
    def raw_path(self) -> str:
        """The raw VPC Lattice request path."""
        return self["raw_path"]

    @property
    def is_base64_encoded(self) -> bool:
        """A boolean flag to indicate if the applicable request payload is Base64-encode"""
        return self["is_base64_encoded"]

    # VPCLattice event has no path field
    # Added here for consistency with the BaseProxyEvent class
    @property
    def path(self) -> str:
        return self["raw_path"]

    @property
    def query_string_parameters(self) -> Dict[str, str]:
        """The request query string parameters."""
        return self["query_string_parameters"]


class vpcLatticeEventV2Identity(DictWrapper):
    @property
    def source_vpc_arn(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext Identity sourceVpcArn"""
        return self.get("sourceVpcArn")

    @property
    def get_type(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext Identity type"""
        return self.get("type")

    @property
    def principal(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext principal"""
        return self.get("principal")

    @property
    def principal_org_id(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext principalOrgID"""
        return self.get("principalOrgID")

    @property
    def session_name(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext sessionName"""
        return self.get("sessionName")

    @property
    def x509_subject_cn(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext X509SubjectCn"""
        return self.get("X509SubjectCn")

    @property
    def x509_issuer_ou(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext X509IssuerOu"""
        return self.get("X509IssuerOu")

    @property
    def x509_san_dns(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext X509SanDns"""
        return self.get("x509SanDns")

    @property
    def x509_san_uri(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext X509SanUri"""
        return self.get("X509SanUri")

    @property
    def x509_san_name_cn(self) -> Optional[str]:
        """The VPC Lattice v2 Event requestContext X509SanNameCn"""
        return self.get("X509SanNameCn")


class vpcLatticeEventV2RequestContext(DictWrapper):
    @property
    def service_network_arn(self) -> str:
        """The VPC Lattice v2 Event requestContext serviceNetworkArn"""
        return self["serviceNetworkArn"]

    @property
    def service_arn(self) -> str:
        """The VPC Lattice v2 Event requestContext serviceArn"""
        return self["serviceArn"]

    @property
    def target_group_arn(self) -> str:
        """The VPC Lattice v2 Event requestContext targetGroupArn"""
        return self["targetGroupArn"]

    @property
    def identity(self) -> vpcLatticeEventV2Identity:
        """The VPC Lattice v2 Event requestContext identity"""
        return vpcLatticeEventV2Identity(self["identity"])

    @property
    def region(self) -> str:
        """The VPC Lattice v2 Event requestContext serviceNetworkArn"""
        return self["region"]

    @property
    def time_epoch(self) -> float:
        """The VPC Lattice v2 Event requestContext timeEpoch"""
        return self["timeEpoch"]


class VPCLatticeEventV2(VPCLatticeEventBase):
    @property
    def version(self) -> str:
        """The VPC Lattice v2 Event version"""
        return self["version"]

    @property
    def is_base64_encoded(self) -> Optional[bool]:
        """A boolean flag to indicate if the applicable request payload is Base64-encode"""
        return self.get("isBase64Encoded")

    @property
    def path(self) -> str:
        """The VPC Lattice v2 Event path"""
        return self["path"]

    @property
    def request_context(self) -> vpcLatticeEventV2RequestContext:
        """he VPC Lattice v2 Event request context."""
        return vpcLatticeEventV2RequestContext(self["requestContext"])

    @property
    def query_string_parameters(self) -> Optional[Dict[str, str]]:
        """The request query string parameters."""
        return self.get("queryStringParameters")



"""Maintenance: We can drop this upon Py3.7 EOL. It's a backport for "location" key to work."""
from __future__ import annotations

import io
import logging
import os
import traceback


def findCaller(stack_info=False, stacklevel=2):  # pragma: no cover
    """
    Find the stack frame of the caller so that we can note the source
    file name, line number and function name.
    """
    f = logging.currentframe()  # noqa: VNE001
    # On some versions of IronPython, currentframe() returns None if
    # IronPython isn't run with -X:Frames.
    if f is None:
        return "(unknown file)", 0, "(unknown function)", None
    while stacklevel > 0:
        next_f = f.f_back
        if next_f is None:
            ## We've got options here.
            ## If we want to use the last (deepest) frame:
            break
            ## If we want to mimic the warnings module:
            # return ("sys", 1, "(unknown function)", None) # noqa: ERA001
            ## If we want to be pedantic:  # noqa: ERA001
            # raise ValueError("call stack is not deep enough") # noqa: ERA001
        f = next_f  # noqa: VNE001
        if not _is_internal_frame(f):
            stacklevel -= 1
    co = f.f_code
    sinfo = None
    if stack_info:
        with io.StringIO() as sio:
            sio.write("Stack (most recent call last):\n")
            traceback.print_stack(f, file=sio)
            sinfo = sio.getvalue()
            if sinfo[-1] == "\n":
                sinfo = sinfo[:-1]
    return co.co_filename, f.f_lineno, co.co_name, sinfo


# The following is based on warnings._is_internal_frame. It makes sure that
# frames of the import mechanism are skipped when logging at module level and
# using a stacklevel value greater than one.
def _is_internal_frame(frame):  # pragma: no cover
    """Signal whether the frame is a CPython or logging module internal."""
    filename = os.path.normcase(frame.f_code.co_filename)
    return filename == logging._srcfile or ("importlib" in filename and "_bootstrap" in filename)




import pytest

from glom import glom, Path, S, A, T, Vars, Val, GlomError, M, SKIP, Let

from glom.core import ROOT
from glom.mutation import PathAssignError

def test_s_scope_assign():
    data = {'a': 1, 'b': [{'c': 2}, {'c': 3}]}
    output = [{'a': 1, 'c': 2}, {'a': 1, 'c': 3}]
    assert glom(data, (S(a=T['a']), ('b', [{'a': S['a'], 'c': 'c'}]))) == output
    assert glom(data, ('b', [{'a': S[ROOT][Val(T)]['a'], 'c': 'c'}])) == output

    with pytest.raises(TypeError):
        S('posarg')
    with pytest.raises(TypeError):
        S()

    assert glom([[1]], (S(v=Vars()), [[A.v.a]], S.v.a)) == 1
    assert glom(1, (S(v={}), A.v['a'], S.v['a'])) == 1
    with pytest.raises(GlomError):
        glom(1, (S(v=1), A.v.a))

    class FailAssign(object):
        def __setattr__(self, name, val):
            raise Exception('nope')

    with pytest.raises(PathAssignError):
        glom(1, (S(v=FailAssign()), Path(A.v, 'a')))

    assert repr(S(a=T.a.b)) == 'S(a=T.a.b)'

    spec = (S(a=T['x']), S.a)
    assert glom({'x': 'y'}, spec) == 'y'

    return


def test_globals():
    assert glom([[1]], ([[A.globals.a]], S.globals.a)) == 1


def test_vars():
    assert glom(1, A.a) == 1  # A should not change the target
    assert glom(1, (A.a, S.a)) == 1
    # check that tuple vars don't "leak" into parent tuple
    assert glom(1, (A.t, Val(2), A.t, S.t)) == 2
    assert glom(1, (A.t, (Val(2), A.t), S.t)) == 1
    let = S(v=Vars({'b': 2}, c=3))
    assert glom(1, (let, A.v.a, S.v.a)) == 1
    with pytest.raises(AttributeError):
        glom(1, (let, S.v.a))  # check that Vars() inside a spec doesn't hold state
    assert glom(1, (let, Path(A, 'v', 'a'), S.v.a)) == 1
    assert glom(1, (let, S.v.b)) == 2
    assert glom(1, (let, S.v.c)) == 3
    assert repr(let) == "S(v=Vars({'b': 2}, c=3))"
    assert repr(Vars(a=1, b=2)) in (
        "Vars(a=1, b=2)", "Vars(b=2, a=1)")
    assert repr(Vars(a=1, b=2).glomit(None, None)) in (
        "ScopeVars({'a': 1, 'b': 2})", "Vars({'b': 2, 'a': 1})")

    assert repr(A.b["c"]) == "A.b['c']"


def test_scoped_vars():
    target = list(range(10)) + list(range(5))

    scope_globals = glom(target, ([A.globals.last], S.globals))
    assert scope_globals.last == 4
    assert dict(scope_globals) == {'last': 4}


def test_max_skip():
    target = list(range(10)) + list(range(5))

    max_spec = (S(max=Vars(max=0)),
                [((M > M(S.max.max)) & A.max.max) | Val(SKIP)],
                 S.max)
    result = glom(target, max_spec)
    assert result.max == 9


def test_let():  # backwards compat 2020-07
    data = {'a': 1, 'b': [{'c': 2}, {'c': 3}]}
    output = [{'a': 1, 'c': 2}, {'a': 1, 'c': 3}]
    assert glom(data, (Let(a='a'), ('b', [{'a': S['a'], 'c': 'c'}]))) == output
    assert glom(data, ('b', [{'a': S[ROOT][Val(T)]['a'], 'c': 'c'}])) == output

    with pytest.raises(TypeError):
        Let('posarg')
    with pytest.raises(TypeError):
        Let()

    assert glom([[1]], (Let(v=Vars()), [[A.v.a]], S.v.a)) == 1
    assert glom(1, (Let(v=lambda t: {}), A.v['a'], S.v['a'])) == 1
    with pytest.raises(GlomError):
        glom(1, (Let(v=lambda t: 1), A.v.a))

    class FailAssign(object):
        def __setattr__(self, name, val):
            raise Exception('nope')

    with pytest.raises(PathAssignError):
        glom(1, (Let(v=lambda t: FailAssign()), Path(A.v, 'a')))

    assert repr(Let(a=T.a.b)) == 'Let(a=T.a.b)'



"""
    Note: This utility is currently in a Non-General Availability (Non-GA) phase and may have limitations.
    Please DON'T USE THIS utility in production environments.
    Keep in mind that when we transition to General Availability (GA), there might be breaking changes introduced.
"""

from aws_lambda_powertools.utilities._data_masking.base import DataMasking

__all__ = [
    "DataMasking",
]



# -*- coding: utf-8 -*-
import os
import re
import sys
import traceback

import pytest

from glom import glom, S, T, GlomError, Switch, Coalesce, Or, Path
from glom.core import format_oneline_trace, format_target_spec_trace, bbrepr, ROOT, LAST_CHILD_SCOPE
from glom.matching import M, MatchError, TypeMatchError, Match


# basic tests:

def test_good_error():
    target = {'data': [0, 1, 2]}

    with pytest.raises(GlomError):
        glom(target, ('data.3'))


def test_error():
    target = {'data': [0, 1, 2]}

    with pytest.raises(GlomError):
        glom(target, ('data', '3'))
    with pytest.raises(GlomError):
        glom(target, ('data', [(T.real, T.bit_length, T.image)]))


def test_pae_api():
    target = {'value': {'data': [0, 1, 2]}}

    with pytest.raises(GlomError) as exc_info:
        glom(target, (T['value'], 'data.3'))

    assert exc_info.value.path == Path('data', '3')
    assert exc_info.value.path.__class__ is Path
    assert exc_info.value.exc.__class__ is IndexError
    assert exc_info.value.part_idx == 1


def test_unfinalized_glomerror_repr():
    assert 'GlomError()' in repr(GlomError())


# trace unit tests:

def test_line_trace():
    stacklifier = ([{'data': S}],)
    scope = glom([1], stacklifier)[0]['data'][ROOT][LAST_CHILD_SCOPE]
    fmtd_stack = format_oneline_trace(scope)
    assert fmtd_stack == '/tuple!list/list<0>/dict!int/S'


# full traceback testing:

def _norm_stack(formatted_stack, exc):

    normalized = []
    for line in formatted_stack.splitlines():
        if line.strip().startswith(u'File'):
            file_name = line.split(u'"')[1]
            short_file_name = os.path.split(file_name.strip(u'"'))[1]
            line = line.replace(file_name, short_file_name)
            line = line.partition(u'line')[0] + u'line ___,' + line.partition(u'line')[2].partition(u',')[2]
        line = line.partition(u'0x')[0]  # scrub memory addresses

        line = line.rstrip()  # trailing whitespace shouldn't matter

        # qualify python2's unqualified error type names
        exc_type_name = exc.__class__.__name__
        if exc_type_name in line:
            mod_name = str(getattr(exc.__class__, '__module__', '') or '')
            exc_type_qual_name = exc_type_name
            if 'builtin' not in mod_name:
                exc_type_qual_name = mod_name + '.' + exc_type_name
            if exc_type_qual_name not in line:
                line = line.replace(exc_type_name, exc_type_qual_name)

        normalized.append(line)

    stack = u"\n".join(normalized) + u'\n'
    stack = stack.replace(u',)', u')')  # py37 likes to do Exception('msg',)
    return stack


def _make_stack(spec, **kwargs):
    target = kwargs.pop('target', [None])
    assert not kwargs
    _orig_some_str = getattr(traceback, '_some_str', None)

    def _debug_some_str(value, *a, **kw):
        # all to debug some CI flakiness
        try:
            return str(value)
        except BaseException as be:  # pragma: no cover
            try:
                print(' !! failed to stringify %s object, got %s' % (type(value).__name__, be))
                traceback.print_exc()
            except:
                print(' !! unable to print trace')
            return '<unprintable %s object got %s>' % (type(value).__name__, be)

    traceback._some_str = _debug_some_str

    try:
        try:
            glom(target, spec)
        except GlomError as e:
            stack = _norm_stack(traceback.format_exc(), e)
    finally:
        traceback._some_str = _orig_some_str
    return stack


# quick way to get a function in this file, which doesn't have a glom
# package file path prefix on it. this prevents the function getting
# removed in the stack flattening.
from boltons.funcutils import FunctionBuilder
fb = FunctionBuilder(name='_raise_exc',
                     body='raise Exception("unique message")',
                     args=['t'])
_raise_exc = fb.get_func()

# NB: if we keep this approach, eventually
# boltons.funcutils.FunctionBuilder will put lines into the linecache,
# and comparisons may break


def test_regular_error_stack():
    actual = _make_stack({'results': [{'value': _raise_exc}]})
    expected = """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.GlomError.wrap(Exception): error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: [None]
 - Spec: {'results': [{'value': <function _raise_exc at
 - Spec: [{'value': <function _raise_exc at
 - Target: None
 - Spec: {'value': <function _raise_exc at
 - Spec: <function _raise_exc at
  File "<boltons.funcutils.FunctionBuilder-0>", line ___, in _raise_exc
Exception: unique message
"""
    # _raise_exc being present in the second-to-last line above tests
    # that errors in user-defined functions result in frames being
    # visible
    assert expected == actual


def test_glom_error_stack():
    # NoneType has not attribute value
    expected = u"""\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.PathAccessError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: [None]
 - Spec: {'results': [{'valué': 'value'}]}
 - Spec: [{'valué': 'value'}]
 - Target: None
 - Spec: {'valué': 'value'}
 - Spec: 'value'
glom.core.PathAccessError: could not access 'value', part 0 of Path('value'), got error: AttributeError("'NoneType' object has no attribute 'value'")
"""
    #import glom.core
    #glom.core.GLOM_DEBUG = True
    actual = _make_stack({'results': [{u'valué': u'value'}]})
    print(actual)

    assert actual == expected


# used by the test below, but at the module level.
# originally added to make stack traces more uniform between 
# python versions (py3 tries to qualify lambdas and
# other functions inside of local scopes.)

def _uses_another_glom():
    try:
        ret = glom(['Nested'], {'internal': ['val']})
    except Exception as exc:
        raise


def _subglom_wrap(t):
    return _uses_another_glom()


def test_glom_error_double_stack():
    actual = _make_stack({'results': [{'value': _subglom_wrap}]})
    expected = """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.PathAccessError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: [None]
 - Spec: {'results': [{'value': <function _subglom_wrap at
 - Spec: [{'value': <function _subglom_wrap at
 - Target: None
 - Spec: {'value': <function _subglom_wrap at
 - Spec: <function _subglom_wrap at
glom.core.PathAccessError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: ['Nested']
 - Spec: {'internal': ['val']}
 - Spec: ['val']
 - Target: 'Nested'
 - Spec: 'val'
glom.core.PathAccessError: could not access 'val', part 0 of Path('val'), got error: AttributeError("'str' object has no attribute 'val'")
"""
    assert expected == actual


def test_long_target_repr():
    import glom as glom_mod
    assert not glom_mod.core.GLOM_DEBUG
    actual = _make_stack(target=[None] * 1000, spec='1001')
    assert '(len=1000)' in actual

    class ObjectWithLongRepr(object):
        def __repr__(self):
            return '<%s %s>' % (self.__class__.__name__, 'w' + ('ooooo' * 250))

    actual = _make_stack(target=ObjectWithLongRepr(), spec='badattr')
    assert '...' in actual
    assert '(len=' not in actual  # no length on a single object


def test_branching_stack():
    # ends-in-branch
    actual = _make_stack(Match(Switch(
        [(1, 1), ('a', 'a'), (T.a, T.a)])))
    expected = """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.matching.MatchError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: [None]
 - Spec: Match(Switch([(1, 1), ('a', 'a'), (T.a, T.a)]))
 + Spec: Switch([(1, 1), ('a', 'a'), (T.a, T.a)])
 |\\ Spec: 1
 |X glom.matching.MatchError: [None] does not match 1
 |\\ Spec: 'a'
 |X glom.matching.MatchError: [None] does not match 'a'
 |\\ Spec: T.a
 |X glom.core.PathAccessError: could not access 'a', part 0 of T.a, got error: AttributeError("'list' object has no attribute 'a'")
glom.matching.MatchError: no matches for target in Switch
"""
    assert actual == expected


def test_midway_branch():
    # midway branch, but then continues
    actual = _make_stack(Match(Switch(
        [(1, 1), ('a', 'a'), ([None], T.a)])))
    expected = """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.PathAccessError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: [None]
 - Spec: Match(Switch([(1, 1), ('a', 'a'), ([None], T.a)]))
 + Spec: Switch([(1, 1), ('a', 'a'), ([None], T.a)])
 |\\ Spec: 1
 |X glom.matching.MatchError: [None] does not match 1
 |\\ Spec: 'a'
 |X glom.matching.MatchError: [None] does not match 'a'
 |\\ Spec: [None]
 || Spec: T.a
glom.core.PathAccessError: could not access 'a', part 0 of T.a, got error: AttributeError("'list' object has no attribute 'a'")
"""
    assert actual == expected
    # branch and another branch
    actual = _make_stack(Match(Switch(
        [(1, 1), ('a', 'a'), ([None], Switch(
            [(1, 1), ('a', 'a'), ([None], T.a)]))])))
    expected = """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.PathAccessError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: [None]
 - Spec: Match(Switch([(1, 1), ('a', 'a'), ([None], Switch([(1, 1), ('a', '...
 + Spec: Switch([(1, 1), ('a', 'a'), ([None], Switch([(1, 1), ('a', 'a'), (...
 |\\ Spec: 1
 |X glom.matching.MatchError: [None] does not match 1
 |\\ Spec: 'a'
 |X glom.matching.MatchError: [None] does not match 'a'
 |\\ Spec: [None]
 |+ Spec: Switch([(1, 1), ('a', 'a'), ([None], T.a)])
 ||\\ Spec: 1
 ||X glom.matching.MatchError: [None] does not match 1
 ||\\ Spec: 'a'
 ||X glom.matching.MatchError: [None] does not match 'a'
 ||\\ Spec: [None]
 ||| Spec: T.a
glom.core.PathAccessError: could not access 'a', part 0 of T.a, got error: AttributeError("'list' object has no attribute 'a'")
"""
    assert actual == expected


def test_partially_failing_branch():
    # what happens when part of an Or() etc fails,
    # but another part succeeds and then an error happens further down?
    assert _make_stack((Or('a', T), Or('b', T), 'c'), target=None) == """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.PathAccessError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: None
 - Spec: (Or('a', T), Or('b', T), 'c')
 - Spec: Or('a', T)
 - Spec: Or('b', T)
 - Spec: 'c'
glom.core.PathAccessError: could not access 'c', part 0 of Path('c'), got error: AttributeError("'NoneType' object has no attribute 'c'")
"""


def test_coalesce_stack():
    val = {'a': {'b': 'c'},  # basic dictionary nesting
       'd': {'e': ['f'],    # list in dictionary
             'g': 'h'},
       'i': [{'j': 'k', 'l': 'm'}],  # list of dictionaries
       'n': 'o'}
    actual = _make_stack(Coalesce('xxx', 'yyy'), target=val)
    expected = """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.CoalesceError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: {'a': {'b': 'c'}, 'd': {'e': ['f'], 'g': 'h'}, 'i': [{'j... (len=4)
 + Spec: Coalesce('xxx', 'yyy')
 |\\ Spec: 'xxx'
 |X glom.core.PathAccessError: could not access 'xxx', part 0 of Path('xxx'), got error: KeyError('xxx')
 |\\ Spec: 'yyy'
 |X glom.core.PathAccessError: could not access 'yyy', part 0 of Path('yyy'), got error: KeyError('yyy')
glom.core.CoalesceError: no valid values found. Tried ('xxx', 'yyy') and got (PathAccessError, PathAccessError) (at path [])
"""
    assert actual == expected


def test_nesting_stack():
    # check behavior when a glom stack is nested via data structure not python call stack
    assert _make_stack(('a', 'b', 'c'), target={'a': {'b': {}}}) == """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.PathAccessError: error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: {'a': {'b': {}}}
 - Spec: ('a', 'b', 'c')
 - Spec: 'a'
 - Target: {'b': {}}
 - Spec: 'b'
 - Target: {}
 - Spec: 'c'
glom.core.PathAccessError: could not access 'c', part 0 of Path('c'), got error: KeyError('c')
"""


ERROR_CLASSES = (
    ValueError, NameError, AttributeError, ZeroDivisionError, SyntaxError, ImportError)

def test_error_types():
    """test that try / except work normally through glom"""
    for err_class in ERROR_CLASSES:
        def err_raise(t):
            raise err_class()
        with pytest.raises(err_class):
            glom(None, err_raise)


def test_fallback():
    """
    test that in cases of weird badly behaved exception types,
    the handler falls back from wrapping to just letting the exception
    through
    """
    class BadExc(Exception):
        def __init__(self, first):
            if not first:
                1/0
            self.first = False
            super(BadExc, self).__init__(self.first)

    bad_exc = BadExc(True)

    def raise_bad(t):
        raise bad_exc

    try:
        glom(None, raise_bad)
    except Exception as e:
        assert e is bad_exc


def test_all_public_errors():
    """test that all errors importable from the top-level glom module
    pass a basic set of standards.

    When adding a new public error type, this test will be fail unless
    that type is also tested below.
    """
    import glom
    import copy

    err_types = [t for t in
                 [getattr(glom, name) for name in dir(glom)]
                 if isinstance(t, type) and issubclass(t, Exception)]
    non_glomerrors = [t for t in err_types if not issubclass(t, glom.GlomError)]
    assert not non_glomerrors, "expected all public exception types to subclass GlomError"

    err_types = sorted([t for t in err_types if not t is glom.GlomError],
                       key=lambda t: t.__name__)

    results = []

    def _test_exc(exc_type, target, spec):
        with pytest.raises(exc_type) as exc_info:
            glom.glom(target, spec)
        results.append((target, spec, exc_info.value))
        return exc_info.value

    _test_exc(glom.CheckError, {}, glom.Check(equal_to=[]))

    _test_exc(glom.FoldError, 2, glom.Flatten())

    _test_exc(glom.BadSpec, range(5), glom.grouping.Group([{T: T}]))

    _test_exc(glom.PathAccessError, {}, 'a.b.c')

    _test_exc(glom.UnregisteredTarget, 'kurt', [glom.T])

    _test_exc(glom.CoalesceError, {}, glom.Coalesce('a', 'b'))

    _test_exc(glom.PathAssignError, object(), glom.Assign('a', 'b'))

    _test_exc(glom.PathDeleteError, object(), glom.Delete('a'))

    _test_exc(MatchError, 1, M == 2)

    _test_exc(TypeMatchError, 1, Match(str))

    for (target, spec, exc) in results:
        assert copy.copy(exc) is not exc
        exc_str = str(exc)
        exc_type_name = exc.__class__.__name__
        assert exc_type_name in exc_str
        assert bbrepr(exc).startswith(exc_type_name)

        assert bbrepr(target)[:80] in exc_str
        assert bbrepr(spec)[:80] in exc_str

    tested_types = [type(exc) for _, _, exc in results]
    untested_types = set(err_types) - set(tested_types)

    assert not untested_types, "did not test all public exception types"


def test_glom_dev_debug():
    with pytest.raises(GlomError) as exc_info:
        glom({'a': 'yesandno'}, 'a.b.c')

    assert ' - Target:' in str(exc_info.value)
    assert 'yesandno' in str(exc_info.value)
    assert len(exc_info.traceback) == 2

    with pytest.raises(GlomError) as exc_info:
        glom({'a': 'yesandno'}, 'a.b.c', glom_debug=True)

    assert ' - Target:' not in str(exc_info.value)
    assert len(exc_info.traceback) > 2


def test_unicode_stack():
    val = {u'resumé': u'beyoncé'}
    stack = _make_stack(target=val, spec=u'a.é.i.o')
    assert 'beyonc' in stack
    assert u'é' in stack


def test_3_11_byte_code_caret():
    """the ~~^ used to mark the current byte code position should be removed"""
    assert _make_stack(target=1, spec=[]) == """\
Traceback (most recent call last):
  File "test_error.py", line ___, in _make_stack
    glom(target, spec)
  File "core.py", line ___, in glom
    raise err
glom.core.GlomError.wrap(IndexError): error raised while processing, details below.
 Target-spec trace (most recent last):
 - Target: 1
 - Spec: []
IndexError: list index out of range
"""



from multiprocessing import AuthenticationError
import re
import requests
from polly import helpers
from polly.auth import Polly
from polly.errors import (
    InvalidParameterException,
    InvalidPathException,
    UnauthorizedException,
    InvalidJobFunctionParameterException,
    extract_json_api_error,
    has_error_message,
    error_handler,
)
from polly.helpers import parseInt

# from polly.constants import MIN_REQUIRED_KEYS_FOR_JOBS, MACHINES_FOR_JOBS
from polly.constants import COMPUTE_ENV_VARIABLE, DATA_LOG_MODES
import polly.http_response_codes as http_codes
from polly import constants as const
from polly.tracking import Track

import pandas as pd
import os
import json


class jobs:

    """
    The polly_jobs class contains functions which can be used to create, cancel and monitor polly jobs.
    Polly CLI jobs can now be initiated, managed and have a status-checked for from Polly Python.
    This lets users run jobs on the Polly cloud infrastructure by scaling computation resources as per need.
    Users can start and stop tasks and even monitor jobs.

    Note:- User must use TOKEN instead of KEYS to work with the job module of polly-python.

    Args:
        token (str): token copy from polly.
    Usage:
        from polly.jobs import jobs

        jobs = jobs(token)
    """

    def __init__(
        self,
        token=None,
        env="",
        default_env="polly",
    ) -> None:
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            COMPUTE_ENV_VARIABLE, default_env, env
        )
        self.session = Polly.get_session(token, env=env)
        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.discover_url = f"https://api.discover.{self.session.env}.elucidata.io"

    # project_id -> workspace_id
    # job_file -> json file
    @Track.track_decorator
    def submit_job(self, project_id: str, job_file=None, job_dict=None) -> pd.DataFrame:
        """
        Submits  a polly cli job in the given workspace.
        A message will be displayed on the status of the operation.
        Job configuration can be provided either as a json file using job_file parameter or
        as a dictionary using job_dict parameter

        Args:
            project_id (str):  workspace id
            job_file (str, optional) : job configuration json file path, defaults to None
            job_dict (dict, optional) : job configuration dictionary, defaults to None
        """
        parameter_check_dict = {}

        parameter_check_dict["project_id"] = project_id
        # neither both job_file and job_dict should be provided, nor both should not be provided.
        # only one of the two should be provided
        if (job_file is None and job_dict is None) or (job_file and job_dict):
            raise InvalidParameterException(
                "One of json_file or json_dict required for the function."
                + "Either both or none have been provided.Please check"
            )
        if job_file:
            parameter_check_dict["job_file"] = job_file
        if job_dict:
            parameter_check_dict["job_dict"] = job_dict

        self._parameter_check_for_jobs(parameter_check_dict)
        if isinstance(project_id, int):
            project_id = str(project_id)

        try:
            if job_file:
                with open(job_file, "r") as jobfile:
                    jobData = json.load(jobfile)
                    # self._validate_job_json(jobData)
            else:
                jobData_json_string = json.dumps(job_dict)
                jobData = json.loads(jobData_json_string)
        except Exception as err:
            raise err

        # TODO: add file validation for json
        jobData = self._add_secret_env_keys_to_jobData(jobData, project_id)
        job_post_data = self._submit_job_to_polly(project_id, jobData)
        submittedProject_df = ""
        if job_post_data:
            submitted_job_id = job_post_data.json().get("data").get("job_id")
            project_id = job_post_data.json().get("data").get("project_id")
            submittedProject = {"Workspace ID": project_id, "Job ID": submitted_job_id}
            submittedProject_df = pd.DataFrame([submittedProject])
        return submittedProject_df

    @Track.track_decorator
    def cancel_job(self, project_id: str, job_id: str):
        """
        cancel a polly job.

        Args:
            project_id (str): workspace id
            job_id (str): job id

        Raises:
            InvalidParameterException
        """
        parameter_check_dict = {}
        parameter_check_dict["project_id"] = project_id
        parameter_check_dict["job_id"] = job_id
        self._parameter_check_for_jobs(parameter_check_dict)
        if isinstance(project_id, int):
            project_id = str(project_id)

        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.jobUrl = f"{self.base_url}/projects/{project_id}/jobs/{job_id}"
        try:
            # successfull -> returns 204
            postData = self.session.delete(self.jobUrl)
            if postData.status_code == 204:
                print("Cancelled job ID " + job_id + " successfully!")
            else:
                # handling this differently because postData.text is a str
                if postData.text:
                    postData_resp_json = json.loads(postData.text)
                else:
                    postData_resp_json = postData.json()
                error = postData_resp_json.get("error")
                if error is None:
                    error = postData_resp_json.get("errors", [])[0]
                if "detail" in error:
                    detail = error.get("detail", "")
                    print("Failed to cancel the job.: " + detail)
                else:
                    print("Failed to cancel the job.")
        except Exception as err:
            print("Failed to cancel the job.")
            raise err

    @Track.track_decorator
    def job_status(self, project_id: str, job_id="", internalCalls=False) -> dict:
        """
        Get the status of a job given the rproject id and job id.
        If no job id given, gives status for all the jobs in the
        provided workspace.

        Args:
            project_id (str): workspace id
            job_id (str): job id

        Returns:
            DataFrame : Contains job id, job name and status sorted as per created timestamp
        """
        parameter_check_dict = {}
        parameter_check_dict["project_id"] = project_id
        if job_id:
            parameter_check_dict["job_id"] = job_id
        self._parameter_check_for_jobs(parameter_check_dict)
        if isinstance(project_id, int):
            project_id = str(project_id)

        sortedJobsTemp = None
        sortedJobs = []
        self.jobUrl = f"{self.base_url}/projects/{project_id}/jobs/{job_id}"

        # TODO: Yogesh : from a job id if we can get the project id. -> improvement
        # TODO: job id can be a optional parameter -> MVP
        try:
            postDatas = self._get_data_for_job_id(self.jobUrl)
            if postDatas.status_code != const.OK:
                self._handle_response_errors(postDatas)
            # list of data
            else:
                sortedJobsTemp = postDatas.json()
                sortedJobs.extend(sortedJobsTemp.get("data"))
                # TODO: can remove this looping with dev testing
                for i in range(len(sortedJobsTemp.get("data"))):
                    if sortedJobsTemp.get("links", "") and sortedJobsTemp.get(
                        "links", ""
                    ).get("next", ""):
                        jobUrl = (
                            f"{self.base_url}{sortedJobsTemp.get('links').get('next')}"
                        )
                        postDatas = self._get_data_for_job_id(jobUrl)
                        if postDatas.status_code != const.OK:
                            self._handle_response_errors(postDatas)
                        sortedJobsTemp = postDatas.json()
                        sortedJobs.extend(sortedJobsTemp.get("data"))
                    else:
                        continue
        except Exception as err:
            print("Not able to get the status of the Job(s)")
            raise err
        sortedJobs = sorted(sortedJobs, key=lambda d: d["attributes"]["created_ts"])
        if internalCalls:
            return sortedJobs
        status_result_df = self._generate_job_status_df(sortedJobs)
        # with pd.option_context(
        #     "display.max_rows", 800, "display.max_columns", 800, "display.width", 1200
        # ):
        #     print(status_result_df)
        return status_result_df

    @Track.track_decorator
    def job_logs(self, project_id: any, job_id: str, mode="all"):
        """
        get logs of job
        Arguments:
            project_id (str): workspace_id
            job_id (str): job_id
        Keyword Arguments:
            mode (str): either 'latest' or 'all' logs (default: {"all"})
        Raises:
            err: RequestException
        """
        try:
            parameter_check_dict = {}
            parameter_check_dict["project_id"] = project_id
            parameter_check_dict["job_id"] = job_id
            parameter_check_dict["mode"] = mode
            self._parameter_check_for_jobs(parameter_check_dict)
            if isinstance(project_id, int):
                project_id = str(project_id)
            returndata = []
            # get the latest log of the job submitted
            if mode == "latest":
                templog = self._get_job_log(project_id, job_id, None, True)
                if templog:
                    returndata.append(templog)
                    log_signed_url = templog.get("log_signed_url")
                    response = requests.get(log_signed_url)
                    error_handler(response)
                    logData = response.text
                    print(logData)
            # get all the logs of the jobs submitted
            elif mode == "all":
                nextToken = None
                templog = self._get_job_log(project_id, job_id, nextToken)
                if templog:
                    returndata.append(templog)
                    log_signed_url = templog.get("log_signed_url")
                    response = requests.get(log_signed_url)
                    error_handler(response)
                    logData = response.text
                    print(logData)
                    nextToken = templog.get("next_token", None)
                while nextToken:
                    templog = self._get_job_log(project_id, job_id, nextToken)
                    if templog:
                        returndata.append(templog)
                        log_signed_url = templog.get("log_signed_url")
                        response = requests.get(log_signed_url)
                        error_handler(response)
                        logData = response.text
                        print(logData)
                        nextToken = templog.get("next_token", None)
        except Exception as err:
            print("Error: Not able to get the log of the job")
            raise err

    def _get_job_log(self, project_id: str, job_id: str, next_token=None, latest=False):
        """
        get the AWS signed url for the job logs.
        Returns response (str) with signed url and next token for job logs

        args:
            project_id (str/int): workspace id
            job_id (str): job id

        **kwargs:
            next_token (str): token string to fetch data (default: {None})
            latest (bool) : get latest log (default: {False})

        """
        logReturn = ""
        try:
            env = self.session.env
            self.base_url = f"https://v2.api.{env}.elucidata.io"
            self.jobUrl = f"{self.base_url}/projects/{project_id}/jobs/{job_id}"
            if next_token:
                self.log_url = f"{self.jobUrl}/logs?next_token={next_token}"
            else:
                self.log_url = f"{self.jobUrl}/logs/?"
            if latest:
                if "next_token" in self.log_url:
                    self.log_url = f"{self.log_url}&show=latest"
                else:
                    self.log_url = f"{self.log_url}?show=latest"
            log_returned = self._get_data_for_job_id(self.log_url)
            log_returned_data = log_returned.json().get("data")
            if not log_returned_data:
                error = log_returned.json().get("error")
                if error is None:
                    error = log_returned.json().get("errors", [])[0]
                if error.get("code") == "resource_not_found":
                    print("Not able to find the logs. It seems to be not generated yet")
                    return
                else:
                    print("Not able to get the logs.")
                    return
            logReturn = log_returned_data[0].get("attributes")
            return logReturn
        except Exception as err:
            print("Not able to get the logs.")
            raise err

    def _is_valid_job_id(self, project_id: str, job_id: str):
        """
        checks if the job id provided is valid
        checks the jobid in the jobs in the workspace

        Arguments:
            project_id (int or str):  workspace id
            job_id (str): job id

        Returns:
            valid (bool): True if job id is valid
        """
        valid = False
        try:
            jobs_list = self.job_status(project_id, internalCalls=True)
            job_id_list = []
            for jobs in jobs_list:
                job_id_list.append(jobs.get("attributes").get("job_id"))
            if job_id in job_id_list:
                valid = True
        except Exception as err:
            raise err
        return valid

    def _handle_response_errors(self, postDatas):
        """
        api to handle the response errors

        Arguments:
            postDatas (json) : api response json

        """
        if postDatas.status_code == http_codes.FORBIDDEN:
            raise AuthenticationError("User access is denied. Please contact admin")
        elif postDatas.status_code == http_codes.UNAUTHORIZED:
            raise UnauthorizedException("User is unauthorized to access this")
        elif postDatas.status_code == http_codes.NOT_FOUND:
            error_title, error_detail = extract_json_api_error(postDatas.text)
            raise Exception(error_detail)
        elif postDatas.status_code == http_codes.GATEWAY_TIMEOUT:
            error_detail = postDatas.get("message", "Request timed out")
        else:
            error_title, error_detail = extract_json_api_error(postDatas.text)
            raise Exception(error_detail)

    def _generate_job_status_df(self, sortedJobs):
        """
        given a list of jobs related info dict this generates a dataframe of the job id, name and status

        Arguments:
            sortedJobs (list): list of job information

        Returns:
            status_result_df (pd.dataframe): dataframe containing job id, name
        """
        job_info_list = []
        if sortedJobs:
            for jobs in sortedJobs:
                job_info = {}
                job_info["id"] = jobs.get("attributes").get("job_id")
                job_info["name"] = jobs.get("attributes").get("job_name")
                job_info["job_state"] = jobs.get("attributes").get("state")
                job_info_list.append(job_info)
            status_result_df = pd.DataFrame(job_info_list)
            status_result_df.columns = ["Job ID", "Job Name", "Job State"]
        else:
            status_result_df = pd.DataFrame(columns=["Job ID", "Job Name", "Job State"])
        return status_result_df

    def _get_data_for_job_id(self, jobUrl):
        """
        given a url, gets the response for the url.
        """
        try:
            if jobUrl:
                self.jobUrl = jobUrl
                postDatas = self.session.get(self.jobUrl)
            else:
                postDatas = self.session.get(self.jobUrl)
        except Exception as err:
            raise err
        return postDatas

    def _add_secret_env_keys_to_jobData(self, jobData: json, project_id: str) -> json:
        """
        add env variables and values to the job data json needed for job run

        Args:
            jobData (json) :  json containing existing variables
            project_id (str):  workspace id

        Returns:
           appended job data json with the secret env variables
        """
        if "secret_env" not in jobData:
            jobData["secret_env"] = {}
        headers = self.session.headers
        if "x-api-key" in headers:
            jobData["secret_env"]["POLLY_API_KEY"] = headers.get("x-api-key")
        jobData["secret_env"]["POLLY_WORKSPACE_ID"] = project_id
        return jobData

    def _submit_job_to_polly(self, project_id: str, job_data: json):
        """
        given a complete job data and workspace id
        calling the api to start a job

        Args:
            project_id (str) : workspace id
            job_data (jsob) : json containing all required variables

        Raises:
            err

        Returns:
            response from the api call
        """
        submitBody = {"data": {"type": "jobs", "attributes": job_data}}

        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.jobUrl = f"{self.base_url}/projects/{project_id}/jobs"
        try:
            postData = self.session.post(self.jobUrl, data=json.dumps(submitBody))
            if postData.status_code != const.CREATED:
                if postData.status_code == http_codes.GATEWAY_TIMEOUT:
                    error_detail = postData.json().get("message", "Request timed out")
                    raise Exception(error_detail)
                if has_error_message(postData):
                    title, detail = extract_json_api_error(postData)
                    raise Exception(title, detail)
        except Exception as err:
            print("Not able to submit job")
            raise err
        return postData

    def _valid_cpu(self, cpu: str) -> bool:
        matches = False
        cpu_regex = "^([+-]?[0-9.]+)([m]*[-+]?[0-9]*)$"
        match = re.fullmatch(cpu_regex, cpu)
        if match is not None:
            matches = True
        return matches

    def _convert_cpu(self, cpu: str):
        cpuRegex = "^[0-9]+m$"
        if re.fullmatch(cpuRegex, cpu):
            cpu_parsed = parseInt(cpu[0:-1]) * 0.001
            return cpu_parsed
        else:
            return parseInt(cpu)

    def _parameter_check_for_jobs(self, paramter_dict: dict):
        for keys, values in paramter_dict.items():
            if keys == "project_id":
                project_id = values
                if not (
                    project_id
                    and (isinstance(project_id, str) or isinstance(project_id, int))
                ):
                    raise InvalidParameterException("project id/workspace id")
            elif keys in ["job_file", "job_dict"]:
                self._check_job_configuration_parameter(keys, values)
            elif keys == "job_id":
                job_id = values
                self._check_job_id(job_id, project_id)
            elif keys == "mode":
                mode = values
                if mode not in DATA_LOG_MODES:
                    raise InvalidParameterException(
                        f"valid mode values are : {DATA_LOG_MODES}"
                    )

    def _check_job_configuration_parameter(
        self, job_config_param_key, job_config_param_value
    ):
        if job_config_param_key == "job_file":
            job_file = job_config_param_value
            if not os.path.exists(job_file):
                raise InvalidPathException(job_file)
        elif job_config_param_key == "job_dict":
            job_dict = job_config_param_value
            if not (isinstance(job_dict, dict)):
                raise InvalidParameterException(
                    "job_dict should be a dictionary datatype"
                )

    def _check_job_id(self, job_id, project_id):
        if not (job_id and (isinstance(job_id, str))):
            raise InvalidParameterException("Missing/invalid datatype for job id")
        if not self._is_valid_job_id(project_id, job_id):
            raise InvalidJobFunctionParameterException("Job id")



"""
Serialization for supporting idempotency
"""
from abc import ABC, abstractmethod
from typing import Any, Dict


class BaseIdempotencySerializer(ABC):
    """
    Abstract Base Class for Idempotency serialization layer, supporting dict operations.
    """

    @abstractmethod
    def to_dict(self, data: Any) -> Dict:
        raise NotImplementedError("Implementation of to_dict is required")

    @abstractmethod
    def from_dict(self, data: Dict) -> Any:
        raise NotImplementedError("Implementation of from_dict is required")


class BaseIdempotencyModelSerializer(BaseIdempotencySerializer):
    """
    Abstract Base Class for Idempotency serialization layer, for using a model as data object representation.
    """

    @classmethod
    @abstractmethod
    def instantiate(cls, model_type: Any) -> BaseIdempotencySerializer:
        """
        Creates an instance of a serializer based on a provided model type.
        In case the model_type is unknown, None will be sent as `model_type`.
        It's on the implementer to verify that:
        - None is handled correctly
        - A model type not matching the expected types is handled

        Parameters
        ----------
        model_type: Any
            The model type to instantiate the class for

        Returns
        -------
        BaseIdempotencySerializer
            Instance of the serializer class
        """
        pass



from aws_lambda_powertools.shared.types import Literal
from aws_lambda_powertools.utilities.data_classes.common import DictWrapper


class SecretsManagerEvent(DictWrapper):
    @property
    def secret_id(self) -> str:
        """SecretId: The secret ARN or identifier"""
        return self["SecretId"]

    @property
    def client_request_token(self) -> str:
        """ClientRequestToken: The ClientRequestToken associated with the secret version"""
        return self["ClientRequestToken"]

    @property
    def version_id(self) -> str:
        """Alias to ClientRequestToken to get token associated to version"""
        return self["ClientRequestToken"]

    @property
    def step(self) -> Literal["createSecret", "setSecret", "testSecret", "finishSecret"]:
        """Step: The rotation step (one of createSecret, setSecret, testSecret, or finishSecret)"""
        return self["Step"]



from __future__ import annotations

import functools
import logging
from abc import ABC, abstractmethod
from typing import Any

from aws_lambda_powertools.metrics.provider import cold_start
from aws_lambda_powertools.shared.types import AnyCallableT
from aws_lambda_powertools.utilities.typing import LambdaContext

logger = logging.getLogger(__name__)


class BaseProvider(ABC):
    """
    Interface to create a metrics provider.

    BaseProvider implements `log_metrics` decorator for every provider as a value add feature.

    Usage:
        1. Inherit from this class.
        2. Implement the required methods specific to your metric provider.
        3. Customize the behavior and functionality of the metric provider in your subclass.
    """

    @abstractmethod
    def add_metric(self, *args: Any, **kwargs: Any) -> Any:
        """
        Abstract method for adding a metric.

        This method must be implemented in subclasses to add a metric and return a combined metrics dictionary.

        Parameters
        ----------
        *args:
            Positional arguments.
        *kwargs:
            Keyword arguments.

        Returns
        ----------
        Dict
            A combined metrics dictionary.

        Raises
        ----------
        NotImplementedError
            This method must be implemented in subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def serialize_metric_set(self, *args: Any, **kwargs: Any) -> Any:
        """
        Abstract method for serialize a metric.

        This method must be implemented in subclasses to add a metric and return a combined metrics dictionary.

        Parameters
        ----------
        *args:
            Positional arguments.
        *kwargs:
            Keyword arguments.

        Returns
        ----------
        Dict
            Serialized metrics

        Raises
        ----------
        NotImplementedError
            This method must be implemented in subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def flush_metrics(self, *args: Any, **kwargs) -> Any:
        """
        Abstract method for flushing a metric.

        This method must be implemented in subclasses to add a metric and return a combined metrics dictionary.

        Parameters
        ----------
        *args:
            Positional arguments.
        *kwargs:
            Keyword arguments.

        Raises
        ----------
        NotImplementedError
            This method must be implemented in subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def clear_metrics(self, *args: Any, **kwargs) -> None:
        """
        Abstract method for clear metric instance.

        This method must be implemented in subclasses to clear the metric instance

        Parameters
        ----------
        *args:
            Positional arguments.
        *kwargs:
            Keyword arguments.

        Raises
        ----------
        NotImplementedError
            This method must be implemented in subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def add_cold_start_metric(self, context: LambdaContext) -> Any:
        """
        Abstract method for clear metric instance.

        This method must be implemented in subclasses to add a metric and return a combined metrics dictionary.

        Parameters
        ----------
        *args:
            Positional arguments.
        *kwargs:
            Keyword arguments.

        Raises
        ----------
        NotImplementedError
            This method must be implemented in subclasses.
        """
        raise NotImplementedError

    def log_metrics(
        self,
        lambda_handler: AnyCallableT | None = None,
        capture_cold_start_metric: bool = False,
        raise_on_empty_metrics: bool = False,
        **kwargs,
    ):
        """Decorator to serialize and publish metrics at the end of a function execution.

        Be aware that the log_metrics **does call* the decorated function (e.g. lambda_handler).

        Example
        -------
        **Lambda function using tracer and metrics decorators**

            from aws_lambda_powertools import Metrics, Tracer

            metrics = Metrics(service="payment")
            tracer = Tracer(service="payment")

            @tracer.capture_lambda_handler
            @metrics.log_metrics
            def handler(event, context):
                    ...

        Parameters
        ----------
        lambda_handler : Callable[[Any, Any], Any], optional
            lambda function handler, by default None
        capture_cold_start_metric : bool, optional
            captures cold start metric, by default False
        raise_on_empty_metrics : bool, optional
            raise exception if no metrics are emitted, by default False
        default_dimensions: Dict[str, str], optional
            metric dimensions as key=value that will always be present

        Raises
        ------
        e
            Propagate error received
        """
        extra_args = {}

        if kwargs.get("default_dimensions"):
            extra_args.update({"default_dimensions": kwargs.get("default_dimensions")})

        if kwargs.get("default_tags"):
            extra_args.update({"default_tags": kwargs.get("default_tags")})

        # If handler is None we've been called with parameters
        # Return a partial function with args filled
        if lambda_handler is None:
            logger.debug("Decorator called with parameters")
            return functools.partial(
                self.log_metrics,
                capture_cold_start_metric=capture_cold_start_metric,
                raise_on_empty_metrics=raise_on_empty_metrics,
                **extra_args,
            )

        @functools.wraps(lambda_handler)
        def decorate(event, context, *args, **kwargs):
            try:
                response = lambda_handler(event, context, *args, **kwargs)
                if capture_cold_start_metric:
                    self._add_cold_start_metric(context=context)
            finally:
                self.flush_metrics(raise_on_empty_metrics=raise_on_empty_metrics)

            return response

        return decorate

    def _add_cold_start_metric(self, context: Any) -> None:
        """
        Add cold start metric

        Parameters
        ----------
        context : Any
            Lambda context
        """
        if not cold_start.is_cold_start:
            return

        logger.debug("Adding cold start metric and function_name dimension")
        self.add_cold_start_metric(context=context)

        cold_start.is_cold_start = False


def reset_cold_start_flag_provider():
    if not cold_start.is_cold_start:
        cold_start.is_cold_start = True



import requests
from polly.__init__ import __version__ as current_version
from polly.session import PollySession
from requests.adapters import HTTPAdapter, Retry
from polly import helpers
from polly import constants as const
from polly.help import example
from packaging import version

link_doc = "https://docs.elucidata.io/OmixAtlas/Polly%20Python.html"
retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])


class UnauthorizedException(Exception):
    """

    :meta private:
    """

    def __str__(self):
        return f"Authorization failed as credentials not found. Please use Polly.auth(token) as shown here  ---- {link_doc}"


class Polly:
    """
    This class for authorization to use Polly on any local or cloud platform. To authenticate usage of Polly, \
the following function can be used.
    """

    default_session = None
    example = classmethod(example)
    auth_flag = False  # Flag that represents session set or not

    @classmethod
    def auth(cls, token, env="", default_env="polly"):
        """
        Function for authorize usage of Polly on terminal or notebook.

        ``Args:``
            |  ``token (str):`` token copy from Polly front-end.


        ``Returns:``
            |  If token is not correct, it will give an error, else it will clear the authentication for user to\
 get started with polly-python.

        ``Error:``
            |  ``UnauthorizedException:`` when the token is expired.

        To use auth function import class Polly.

        .. code::


                from polly.auth import Polly
                Polly.auth(token)
        """
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            const.COMPUTE_ENV_VARIABLE, default_env, env
        )
        cls.default_session = PollySession(token, env=env)
        cls.default_session.mount(
            "https://",
            HTTPAdapter(pool_connections=100, pool_maxsize=100, max_retries=retries),
        )
        cls._version_check()

    @classmethod
    def get_session(cls, token=None, env="polly"):
        """
        Function to get session from polly.

        ``Args:``
            |  ``token (str):`` token copy from polly.

        ``Returns:``
            |  if token is not satisfied it will throw UnauthorizedException.
            |  else it will return a polly.session object.

        ``Error:``
            |  ``UnauthorizedException:`` when the token is expired.

        To use get_sesion function import class Polly.


        .. code::


                from polly.auth import Polly
                session = Polly.get_session(token)

        """
        if token:
            cls.auth(token, env=env)
        else:
            if not cls.default_session:
                raise UnauthorizedException
        # base_url = f"https://v2.api.{cls.default_session.env}.elucidata.io"
        base_url_auth = f"https://apis.{cls.default_session.env}.elucidata.io/auth"
        cls.default_session.user_details = helpers.get_user_details(
            cls.default_session, base_url_auth
        )
        return cls.default_session

    @classmethod
    def _version_check(cls):
        # check if session is already set, if not then only execute
        if not cls.auth_flag:
            polly_python_latest_version_url = const.POLLY_PYTHON_LATEST_VERSION_FILE
            # GET Request to "polly_python_latest_version.txt" in github for fetching info about polly-py release versions
            try:
                response = requests.get(polly_python_latest_version_url)
                data = response.text
                # removing new line character and character v from the text
                latest_version = data.replace("\n", "").replace("v", "")
                if version.parse(latest_version) > version.parse(current_version):
                    print(
                        f"You're currently using an outdated version of polly-python '{current_version}'. \
Please update using the command \
'pip install polly-python=={latest_version}' to upgrade to the newest version '{latest_version}'"
                    )
            except Exception:
                # Exception to not interrupt the session build
                pass
            # set flag to true which means session has been set once
            cls.auth_flag = True
